
#include <bx.h>

double distance (int x1, int y1, int x2, int y2)
{
   double dx= x1 - x2;
   double dy= y1 - y2;
   return sqrt (dx*dx + dy*dy); 
}

int nearest_point (int xm, int ym, 
                   int n, int x[], int y[])
{
  int k, min;
  double min_dist= INT_MAX;
  for (k= 0; k < n; k++) {
    double dist= distance (xm, ym, x[k], y[k]);
    if (dist < min_dist) { min= k; min_dist= dist; }
  }
  return min;
}

void update_shapes (int xm, int ym, int k, int x[], int y[])
{
  double alpha= 0.1;
  int xk= (x[k]*(1-alpha) + xm*alpha);
  int yk= (y[k]*(1-alpha) + ym*alpha);
  if (xk == x[k] && yk == y[k]) { x[k]= xm; y[k]= ym; }
  else                          { x[k]= xk; y[k]= yk; } 
}

void draw_shapes (bx_window window, int n, int x[], int y[], int mode,
                  int xm, int ym, int k)
{
  int xcurve [BX_CURVEI_LENGTH];
  int ycurve [BX_CURVEI_LENGTH];
  int len, filled, zoom;
  static int xt0= 0, yt0= 0;

  xt0++; yt0++;
  bx_set_tile_origin (xt0, yt0);

  switch (mode) {
  case 0: len= bx_points_to_curve  (n, x, y, 1    , xcurve, ycurve);break;
  case 1: len= bx_points_to_curve2 (n, x, y, 1    , xcurve, ycurve);break;
  case 2: len= bx_points_to_curvei (n, x, y, 1,0.5, xcurve, ycurve);break;
  }
  bx_set_line_width (1);
 
  bx_set_fill_rule (BX_EVEN_ODD_RULE);
  bx_set_line_style (BX_DASH_LINE);
  bx_set_color (bx_pale_gray ());

  bx_draw_polygon (window, n, x, y, BX_CLOSED, BX_FILLED);
  bx_set_color (bx_pink ()); 
  bx_set_fill_style (BX_TILED_FILL);
  bx_draw_polygon (window, len, xcurve, ycurve, BX_CLOSED, BX_FILLED);
   bx_set_fill_style (BX_SOLID_FILL);
  bx_set_color (bx_black ());
  bx_draw_polygon (window, len, xcurve, ycurve, BX_CLOSED, BX_NOT_FILLED);
  bx_set_color (bx_black ());
  bx_draw_polygon (window, n, x, y, BX_CLOSED, BX_NOT_FILLED);
  bx_set_color (bx_black ());
  bx_draw_line (window, xm, ym, x[k], y[k]);
  bx_set_line_style (BX_SOLID_LINE);
  bx_draw_circle (window, x[k]-5, y[k]-5, 10, 10, 0);
  if (bx_on_polyline (xm, ym, len, xcurve, ycurve, BX_CLOSED))
    bx_set_color (bx_violet ());
  filled= bx_in_polygon (xm, ym, len, xcurve, ycurve, BX_EVEN_ODD_RULE); 
  zoom= 1+ bx_in_polygon (xm, ym, n, x, y, BX_EVEN_ODD_RULE);
  bx_draw_circle (window, xm  -5*zoom, ym  -5*zoom, 
                  10*zoom, 10*zoom, filled);  
}

int main (void)
{
  bx_window window;
  unsigned w= 500, h= 500;
# define NB_POINTS 10
  int x[NB_POINTS]= { 10 , 240, 250, 260, 490, 
                      490, 260, 250, 240, 10 };
  int y[NB_POINTS]= { 10 , 10 , 240, 10 , 10 ,
                      260, 260, 490, 260, 260};
 
  int b1_pressed= 0, b3_pressed= 0, mode= 0, k= -1;
  bx_image img;

  bx_init ();

  img= bx_load_image (1, "./APE/fins-5.png");
 
  bx_set_tile_image (img);

  window= bx_create_window ("random curves", 10, 10, w, h);
  srand (time (0));

  bx_set_fill_rule (1);
  for (;;) {
    bx_mouse m= bx_read_mouse (window);
    int xm= bx_mouse_x (m);
    int ym= bx_mouse_y (m);
    if (! bx_mouse_b1 (m)) { 
      k= nearest_point (xm, ym, NB_POINTS, x, y);
      b1_pressed= 0;
    } else if (! b1_pressed) {
      k= nearest_point (xm, ym, NB_POINTS, x, y);
      b1_pressed= 1;
    }
    if (b1_pressed &&
        xm > 0 && xm < (int) w && 
        ym > 0 && ym < (int) h) {
      update_shapes (xm, ym, k, x, y);
    }
    if (! bx_mouse_b3 (m)) { b3_pressed= 0; }
    else if (! b3_pressed) { b3_pressed= 1; mode= (mode+1)%3 ; }          

    bx_clear_canvas (window, bx_white ());
    draw_shapes (window, NB_POINTS, x, y, mode, xm, ym, k);
    bx_show_canvas (window, 10);
  }
  return 0;
}
