
#include "bx.h"

#ifdef BX_AUDIO 
#  include <SDL_audio.h>
#endif


static MyX myx;

struct bx_window_stuff {
  Window window;
/* canvas double-buffer */
  Pixmap back, front; 
/* canvas geometry */
  int x, y;
  unsigned width, height; 
/* window geometry */
  MyX_Geometry geom;
  int win_x, win_y;
  unsigned win_width, win_height; 

  MyX_Handlers  handlers;  
  unsigned long eventMask;
  bx_mouse_function on_mouse_move;
  bx_mouse_function on_mouse_press, on_mouse_release;
  bx_key_function   on_key_press, on_key_release;

};

#define BX_MAX_DIR_BUFFER  100000L
#define BX_MAX_DIR_ENTRIES  10000

struct bx_dir_stuff {
  char buffer    [BX_MAX_DIR_BUFFER ]; 
  char * entries [BX_MAX_DIR_ENTRIES];
  int nb_entries;
  long nb_chars;
};

/*-----------------------------------------------------------------------------
 * 
 *-----------------------------------------------------------------------------
 */


int 
bx_on_io_error (Display * display UNUSED)
{
  fprintf (stderr, "bx to user: good bye !\n");
  exit (0);
}

void
bx_init (void) 
{
  if (-1 == MyX_Init (& myx, "")) {
    fprintf (stderr, "bx_init(): failed.\n");
    exit (1);
  }
  MyX_IgnoreAutoRepeat (& myx, 1);
  XSetIOErrorHandler (bx_on_io_error);
}

void
bx_clean (void)
{
  MyX_Clean (& myx);
}

/*-----------------------------------------------------------------------------
 * handlers interne de MyX
 *-----------------------------------------------------------------------------
 */

static void
Window_KeyPress (MyX * myx UNUSED, XEvent * event, void * data) 
{
  bx_window win;
  win.stuff= data;
  if (win.stuff->on_key_press) {
    unsigned char keycode= event->xkey.keycode;
    win.stuff->on_key_press (win, keycode);
  }
}

static void
Window_KeyRelease (MyX * myx UNUSED, XEvent * event, void * data) 
{
  bx_window win;
  win.stuff= data;
  if (win.stuff->on_key_release) {
    unsigned char keycode= event->xkey.keycode; 
    win.stuff->on_key_release (win, keycode);
  }
}

static unsigned
Button_To_ButtonMask (int button) {
  switch (button) {
  case 1: return Button1Mask;
  case 2: return Button2Mask;
  case 3: return Button3Mask;
  case 4: return Button4Mask;
  case 5: return Button5Mask;
  default: return 0;
  }
}

static void
Window_ButtonPress (MyX * myx UNUSED, XEvent * event, void * data) 
{
  bx_window win;
  win.stuff= data;
  if (win.stuff->on_mouse_press) {
    bx_mouse mouse;
    mouse.xRel= event->xbutton.x; mouse.xAbs= event->xbutton.x_root;
    mouse.yRel= event->xbutton.y; mouse.yAbs= event->xbutton.y_root;  
    mouse.x=  mouse.xRel - bx_canvas_x (win); 
    mouse.y=  mouse.yRel - bx_canvas_y (win);
    mouse.root= event->xbutton.root;
    mouse.child= event->xbutton.window;
    mouse.keyMask= event->xbutton.state;
    mouse.keyMask |= Button_To_ButtonMask ( event->xbutton.button);
    win.stuff->on_mouse_press (win, mouse);
  }
}

static void
Window_ButtonRelease (MyX * myx UNUSED, XEvent * event UNUSED, void * data) 
{
  bx_window win;
  win.stuff= data;
  if (win.stuff->on_mouse_release) {
    bx_mouse mouse;
    mouse.xRel= event->xbutton.x; mouse.xAbs= event->xbutton.x_root;
    mouse.yRel= event->xbutton.y; mouse.yAbs= event->xbutton.y_root;  
    mouse.x=  mouse.xRel - bx_canvas_x (win); 
    mouse.y=  mouse.yRel - bx_canvas_y (win);
    mouse.root= event->xbutton.root;
    mouse.child= event->xbutton.window;
    mouse.keyMask= event->xbutton.state;
    mouse.keyMask |= Button_To_ButtonMask ( event->xbutton.button);
    win.stuff->on_mouse_release (win, mouse);
  }
}


static void
Window_Motion (MyX * myx UNUSED, XEvent * event UNUSED, void * data) 
{
  bx_window win;
  win.stuff= data;
  if (win.stuff->on_mouse_move) {
    bx_mouse mouse;
    mouse.xRel= event->xbutton.x; mouse.xAbs= event->xbutton.x_root;
    mouse.yRel= event->xbutton.y; mouse.yAbs= event->xbutton.y_root;  
    mouse.x=  mouse.xRel - bx_canvas_x (win); 
    mouse.y=  mouse.yRel - bx_canvas_y (win);
    mouse.root= event->xbutton.root;
    mouse.child= event->xbutton.window;
    mouse.keyMask= event->xbutton.state;
    win.stuff->on_mouse_move (win, mouse);
  }
}

/*-----------------------------------------------------------------------------
 * 
 *-----------------------------------------------------------------------------
 */

static void
bx_update_canvas_geometry (bx_window win)
{
  win.stuff->x= ((int)win.stuff->win_width  - (int)win.stuff->width )/ 2;
  win.stuff->y= ((int)win.stuff->win_height - (int)win.stuff->height)/ 2;
}

static void
Window_ConfigureNotify (MyX * myx UNUSED, XEvent * event, void * data) 
{
  bx_window win;
  win.stuff= data;
  win.stuff->win_x     = event->xconfigure.x;
  win.stuff->win_y     = event->xconfigure.y;
  win.stuff->win_width = event->xconfigure.width;
  win.stuff->win_height= event->xconfigure.height;
  bx_update_canvas_geometry (win);
}


static void
Window_Expose (MyX * myx UNUSED, XEvent * event, void * data) 
{
  bx_window win;
  if (event->xexpose.count > 0) return; 
  win.stuff= data;
  bx_update_canvas (win);
}

void 
bx_destroy_window (bx_window window)
{
  XDestroyWindow (myx.display, window.stuff->window);
}

/*-----------------------------------------------------------------------------
 * 
 *-----------------------------------------------------------------------------
 */

bx_window 
bx_create_window (char const name [], int x, int y, unsigned w, unsigned h)
{
  bx_window win;
  MyX_Geometry geom;
  
  win.stuff= malloc (sizeof * win.stuff);
  if (! win.stuff) {
    fprintf (stderr, "bx_create_window () failed: no memory.\n");
    exit (1);
  }
  win.stuff->width = w;
  win.stuff->height= h;

  MyX_InitHandlers (& win.stuff->handlers);
  MyX_SetHandler (& win.stuff->handlers, Expose, Window_Expose); 
  MyX_SetHandler (& win.stuff->handlers, ConfigureNotify, 
                  Window_ConfigureNotify); 

  win.stuff->window=  MyX_CreateWindow (& myx, & win.stuff->handlers, 
                                        win.stuff, myx.root, 
                                        x, y, w, h);
 
  /* .... begin paranoiac mode */
  bx_move_window   (win, x, y); 
  bx_resize_window (win, w, h); 
  bx_rename_window (win, name);
  MyX_GetGeometry (& myx, win.stuff->window, & geom);
  win.stuff->win_x     = geom.x;
  win.stuff->win_y     = geom.x;
  win.stuff->win_width = geom.w;
  win.stuff->win_height= geom.h;
  bx_update_canvas_geometry (win);
  /* .... end paranoiac mode */


  if (0) {
    Atom wmdelwin = XInternAtom (myx.display, "WM_DELETE_WINDOW", False);
    XSetWMProtocols (myx.display, win.stuff->window, & wmdelwin, 1);
  }
  win.stuff->eventMask= ExposureMask | StructureNotifyMask; 
  XSelectInput (myx.display, win.stuff->window, win.stuff->eventMask);
 
  win.stuff->on_mouse_press  = NULL;
  win.stuff->on_mouse_release= NULL;
  win.stuff->on_mouse_move   = NULL;
  win.stuff->on_key_press    = NULL;
  win.stuff->on_key_release  = NULL;
 
  win.stuff->back = XCreatePixmap (myx.display, myx.root, w, h, myx.depth);
  win.stuff->front= XCreatePixmap (myx.display, myx.root, w, h, myx.depth);

  bx_clear_canvas (win, bx_white ());
  bx_set_color (bx_black ());

  XMapRaised  (myx.display, win.stuff->window);
 
  bx_show_canvas (win, 0);

  return win;
}

void
bx_move_window (bx_window window, int x, int y)
{
  XMoveWindow (myx.display, window.stuff->window, x, y);
}

void
bx_resize_window (bx_window window, unsigned w, unsigned h)
{
  XResizeWindow (myx.display, window.stuff->window, w, h);
}

void
bx_rename_window (bx_window window, char const name [])
{
  XStoreName (myx.display, window.stuff->window, name);
}

char *
bx_window_name (bx_window window)
{
  static char name [1024];
  char * fetched;
  XFetchName (myx.display, window.stuff->window, & fetched);
  strncpy (name, fetched, 1024);
  name [1023]= '\0';
  XFree (fetched);
  return name;
}

/*-----------------------------------------------------------------------------
 *
 *-----------------------------------------------------------------------------
 */

unsigned bx_canvas_width  (bx_window win) { return win.stuff->width;  }
unsigned bx_canvas_height (bx_window win) { return win.stuff->height; }

int bx_canvas_x (bx_window w) { return w.stuff->x; }
int bx_canvas_y (bx_window w) { return w.stuff->y; }

unsigned bx_window_width  (bx_window w) { return w.stuff->win_width;  }
unsigned bx_window_height (bx_window w) { return w.stuff->win_height; }

int bx_window_x (bx_window w) { return w.stuff->win_x; }
int bx_window_y (bx_window w) { return w.stuff->win_y; }

/*-----------------------------------------------------------------------------
 * abonnement aux evenements souris et clavier
 *-----------------------------------------------------------------------------
 */

void 
bx_on_key_press (bx_window win, bx_key_function function)
{
  MyX_SetHandler (& win.stuff->handlers, KeyPress, Window_KeyPress);
  win.stuff->on_key_press= function;
  win.stuff->eventMask |= (KeyPressMask);
  XSelectInput (myx.display, win.stuff->window, win.stuff->eventMask);
}

void 
bx_on_key_release (bx_window win, bx_key_function function)
{
  MyX_SetHandler (& win.stuff->handlers, KeyRelease, Window_KeyRelease);
  win.stuff->on_key_release= function;
  win.stuff->eventMask |= (KeyReleaseMask);
  XSelectInput (myx.display, win.stuff->window, win.stuff->eventMask);
}

void 
bx_on_mouse_press  (bx_window win, bx_mouse_function function)
{
  MyX_SetHandler (& win.stuff->handlers, ButtonPress, Window_ButtonPress);
  win.stuff->on_mouse_press= function;
  win.stuff->eventMask |= (ButtonPressMask);
  XSelectInput (myx.display, win.stuff->window, win.stuff->eventMask);
}
void 
bx_on_mouse_release  (bx_window win, bx_mouse_function function)
{
  MyX_SetHandler (& win.stuff->handlers, ButtonRelease, Window_ButtonRelease);
  win.stuff->on_mouse_release= function;
  win.stuff->eventMask |= (ButtonReleaseMask);
  XSelectInput (myx.display, win.stuff->window, win.stuff->eventMask);
}

void 
bx_on_mouse_move (bx_window win, bx_mouse_function function)
{
  MyX_SetHandler (& win.stuff->handlers, MotionNotify, Window_Motion);
  win.stuff->on_mouse_move= function;
  win.stuff->eventMask |= (PointerMotionMask);
  XSelectInput (myx.display, win.stuff->window, win.stuff->eventMask);
}

/*-----------------------------------------------------------------------------
 *
 *-----------------------------------------------------------------------------
 */

void 
bx_loop (void) 
{
  MyX_EventLoop (& myx);
}

/*-----------------------------------------------------------------------------
 *  XDrawArc / XFillArc, fully closed
 *-----------------------------------------------------------------------------
 */

static void
bx_draw_circle_aux (Drawable d, GC gc, int x, int y, 
                    unsigned w, unsigned h, bx_fill filled)
{
  if (filled) 
    XFillArc (myx.display, d, gc, x, y, w, h, 0, 360*64);
  else        
    XDrawArc (myx.display, d, gc, x, y, w,   h, 0, 360*64);
}

void
bx_draw_circle (bx_window win, int x, int y, 
                unsigned w, unsigned h, bx_fill filled)
{
  bx_draw_circle_aux (win.stuff->back, myx.gc, 
                      x, y, w, h, filled);
}

/*-----------------------------------------------------------------------------
 * XDrawArc / XFillArc, generic
 *-----------------------------------------------------------------------------
 */


static void 
bx_draw_arc_aux (Drawable d, GC gc, int x, int y, 
                 unsigned w, unsigned h, 
                 double a, double da, bx_fill filled)
{
  if (filled) XFillArc (myx.display,d,gc, x, y, w, h, a*64,da*64);
  else        XDrawArc (myx.display,d,gc, x, y, w, h, a*64,da*64);
}

void
bx_draw_arc (bx_window win, int x, int y, 
             unsigned w, unsigned h, 
             double a, double da, bx_fill filled)
{
  bx_draw_arc_aux (win.stuff->back, myx.gc, 
                   x, y, w, h, a, da, filled);
}

/*-----------------------------------------------------------------------------
 * XDrawPoint
 *-----------------------------------------------------------------------------
 */


static void 
bx_draw_point_aux(Drawable d, GC gc, int x, int y)
{
  XDrawPoint (myx.display, d, gc, x, y);
}

void 
bx_draw_point(bx_window win, int x, int y)
{
  bx_draw_point_aux (win.stuff->back, myx.gc, x, y);
}

/*-----------------------------------------------------------------------------
 * XDrawLine
 *-----------------------------------------------------------------------------
 */

static void
bx_draw_line_aux (Drawable d, GC gc, int x1, int y1, int x2, int y2)
{
  XDrawLine (myx.display, d, gc, x1, y1, x2, y2);
}

static void
bx_draw_line_sys (bx_window win, int x1, int y1, int x2, int y2)
{
  bx_draw_line_aux (win.stuff->window, myx.gc_sys, x1, y1, x2, y2);
}

void
bx_draw_line (bx_window win, int x1, int y1, int x2, int y2)
{
  bx_draw_line_aux (win.stuff->back, myx.gc, x1, y1, x2, y2);
}

/*-----------------------------------------------------------------------------
 * XDrawRectangle / XFillRectangle  
 *-----------------------------------------------------------------------------
 */

static void
bx_draw_box_aux (Drawable d, GC gc, int x, int y, 
                 unsigned w, unsigned h, bx_fill filled)
{
  if (filled) XFillRectangle (myx.display, d, gc, x, y, w, h);
  else        XDrawRectangle (myx.display, d, gc, x, y, w, h);
}

static void
bx_draw_box_sys (bx_window win, int x, int y,
                 unsigned w, unsigned h, bx_fill filled)
{
  bx_draw_box_aux (win.stuff->window, myx.gc_sys, x, y, w, h, filled);
}  

void
bx_draw_box (bx_window win, int x, int y,
             unsigned w, unsigned h, bx_fill filled)
{
  bx_draw_box_aux (win.stuff->back, myx.gc, x, y, w, h, filled);
}  

/*-----------------------------------------------------------------------------
 * XDrawLines / XFillPolygon
 *-----------------------------------------------------------------------------
 */

void
bx_draw_polygon_aux (Drawable d, GC gc, int n, int x[], int y[], 
                     bx_closure closed, bx_fill filled)
{
  XPoint points [BX_MAX_POLYGON_POINTS+1];
  int k;
  if (n > BX_MAX_POLYGON_POINTS) {
    fprintf (stderr, "bx_draw_polygon() failed: "
             "number of points > %d.\n", BX_MAX_POLYGON_POINTS);
    exit (1);
  }
  for (k= 0; k < n; k++) {
    points[k].x= x[k];
    points[k].y= y[k];
  }
  if (closed) {
    points [n].x= x [0];
    points [n].y= y [0];
    n++;
  }
  if (filled) 
    XFillPolygon (myx.display, d, gc, points, n, Complex, CoordModeOrigin);
  else        
    XDrawLines (myx.display, d, gc, points, n, CoordModeOrigin);
}

void
bx_draw_polygon (bx_window win, int n, int x[], int y[], 
                 bx_closure closed, bx_fill filled)
{
  bx_draw_polygon_aux (win.stuff->back, myx.gc, n, x, y, 
                       closed, filled); 
}

/*-----------------------------------------------------------------------------
 * point is in polygon ?
 *-----------------------------------------------------------------------------
 */

typedef struct bx_homogen { double x, y, z; } bx_homogen;

bx_homogen 
bx_homogen_point ( double x, double y)
{
  bx_homogen h;
  h.x= x; 
  h.y= y; 
  h.z= 1.0;
  return h;
}

bx_homogen 
bx_homogen_vector (double x, double y)
{
  bx_homogen h;
  h.x= x; 
  h.y= y; 
  h.z= 0.0;
  return h;
}

bx_homogen 
bx_homogen_line (double a, double b, double c)
{
  bx_homogen h;
  h.x= a; 
  h.y= b; 
  h.z= c;
  return h;
}

bx_homogen 
bx_homogen_crossproduct (bx_homogen h1, bx_homogen h2)
{
  bx_homogen h;
  h.x= h1.y * h2.z - h1.z * h2.y;
  h.y= h1.z * h2.x - h1.x * h2.z;
  h.z= h1.x * h2.y - h1.y * h2.x;
  return h;
}

double
bx_homogen_dotproduct (bx_homogen h1, bx_homogen h2)
{
  return h1.x * h2.x 
    +    h1.y * h2.y 
    +    h1.z * h2.z;
}

#define BX_HOMOGEN_EPSILON  0.000001

bx_homogen 
bx_homogen_normalize (bx_homogen h)
{
  if (fabs (h.z) < BX_HOMOGEN_EPSILON) 
    return h;
  h.x /= h.z;
  h.y /= h.z;
  h.z= 1.0;
  return h;
}

bx_homogen 
bx_homogen_segment (bx_homogen point1, bx_homogen point2)
{
  return bx_homogen_crossproduct (point1, point2);
}


bx_homogen 
bx_homogen_intersect (bx_homogen line1, bx_homogen line2)
{
  bx_homogen inter= bx_homogen_crossproduct (line1, line2);
  return bx_homogen_normalize (inter);
}

int 
bx_on_polyline (int x, int y, 
                int n, int xpoly[], int ypoly[], 
                bx_closure closed)
{
  int k, j, xmin, xmax, ymin, ymax;
  bx_homogen point, point0, point1, line;
  double test;
  point= bx_homogen_point (x, y);
  for (k= 0, j= 1; k < n; k++, j++) {
    if (j == n) {
      if (closed) j= 0;
      else        break;
    }
    if (ypoly[k] < ypoly[j]) { ymin= ypoly[k]; ymax= ypoly[j]; }
    else                     { ymax= ypoly[k]; ymin= ypoly[j]; }
    if (y < ymin || y > ymax) continue;
    if (xpoly[k] < xpoly[j]) { xmin= xpoly[k]; xmax= xpoly[j]; }
    else                     { xmax= xpoly[k]; xmin= xpoly[j]; }
    if (x < xmin || x > xmax) continue;
    point0= bx_homogen_point (xpoly[k], ypoly[k]);
    point1= bx_homogen_point (xpoly[j], ypoly[j]);
    line= bx_homogen_segment (point0, point1);
    test= bx_homogen_dotproduct (point, line);
    if (test < BX_HOMOGEN_EPSILON) return 1;
  }
  return 0;
}

int 
bx_in_polygon (int x, int y, 
               int n, int xpoly[], int ypoly[],
               bx_fill_rule fill_rule)
{
  int k, j, ymin, ymax, dir, count= 0;
  bx_homogen point0, point1, line, inter, scan;
  scan= bx_homogen_line (0, 1.0, 0.5-y);
  for (k= 0, j= 1; k < n; k++, j++) {
    if (j == n) j= 0;
    point0= bx_homogen_point (xpoly[k], ypoly[k]);
    point1= bx_homogen_point (xpoly[j], ypoly[j]);
    line  = bx_homogen_segment (point0, point1);
    inter = bx_homogen_intersect (line, scan);
    if (fabs (inter.z) < BX_HOMOGEN_EPSILON) continue;
    if (inter.x < x) continue;
    if (ypoly[k] < ypoly[j]) { ymin= ypoly[k]; ymax= ypoly[j]; dir= 1; }
    else                     { ymax= ypoly[k]; ymin= ypoly[j]; dir=-1; }
    if (inter.y < ymin) continue;
    if (inter.y > ymax) continue;
    count += dir; 
  }
  return (fill_rule == BX_EVEN_ODD_RULE) ? 
    (count %  2 != 0) : 
    (count      != 0);
}


/*-----------------------------------------------------------------------------
 * XDrawString, simple
 *-----------------------------------------------------------------------------
 */

static  void
bx_draw_text_aux (Drawable d, GC gc, int x, int y, char text [])
{
  int length= strlen (text);
  XDrawString (myx.display, d, gc, x, y, text, length);
}

static  void
bx_draw_text_sys (bx_window win, int x, int y, char text [])
{
  bx_draw_text_aux (win.stuff->window, myx.gc_sys, x, y, text);
}

 void
bx_draw_text (bx_window win, int x, int y, char text [])
{
  bx_draw_text_aux (win.stuff->back, myx.gc, x, y, text);
}

/*-----------------------------------------------------------------------------
 * XDrawString, variadic a la printf
 *-----------------------------------------------------------------------------
 */

#define BX_MAX_BUF 2048

static void
bx_draw_text_sys_f (bx_window win, int x, int y, char format [], ...)
{
  char buffer [BX_MAX_BUF];
  va_list args;
  va_start (args, format);
  vsprintf (buffer, format, args);
  va_end (args);
  bx_draw_text_aux (win.stuff->window, myx.gc_sys, x, y, buffer);
}

void
bx_draw_text_f (bx_window win, int x, int y, char format [], ...)
{
  char buffer [BX_MAX_BUF];
  va_list args;
  va_start (args, format);
  vsprintf (buffer, format, args);
  va_end (args);
  bx_draw_text_aux (win.stuff->back, myx.gc, x, y, buffer);
}

#undef BX_MAX_BUF


/*-----------------------------------------------------------------------------
 * bezier
 *-----------------------------------------------------------------------------
 */


#define BX_MIDPOINT_XY(M,AX,AY,BX,BY){ M.x=(   AX+BX)/2.; M.y=(   AY+BY)/2.; }
#define BX_3RDPOINT_XY(M,AX,AY,BX,BY){ M.x=(2.*AX+BX)/3.; M.y=(2.*AY+BY)/3.; }

#define BX_MIDPOINT(M,A,B) { M.x=(   A.x+B.x)/2.; M.y=(   A.y+B.y)/2.; }
#define BX_3RDPOINT(M,A,B) { M.x=(2.*A.x+B.x)/3.; M.y=(2.*A.y+B.y)/3.; }

typedef struct bx_point { double x, y; } bx_point;


/*-----------------------------------------------------------------------------
 * 
 *-----------------------------------------------------------------------------
 */

int
bx_points_to_bezier (int x0, int y0, 
                     int x1, int y1, 
                     int x2, int y2, 
                     int x3, int y3,
                     int x_res [], int y_res [])
{
  bx_point a[BX_BEZIER_LENGTH], * src= a;
  bx_point b[BX_BEZIER_LENGTH], * dst= b; 
  int max_src, k_src, k_dst, k, level;

  a[0].x = x0; a[1].x = x1; a[2].x = x2; a[3].x = x3; b[0].x= x0;
  a[0].y = y0; a[1].y = y1; a[2].y = y2; a[3].y = y3; b[0].y= y0;

  for(level= 0, max_src= 3; level < BX_BEZIER_LEVEL; level++) {
    for (k_src= k_dst= 0; k_src < max_src; k_src+= 3, k_dst+= 6) {
      bx_point v;
      BX_MIDPOINT (v           , src[k_src+1], src[k_src+2]);
      BX_MIDPOINT (dst[k_dst+1], src[k_src+0], src[k_src+1]);
      BX_MIDPOINT (dst[k_dst+5], src[k_src+2], src[k_src+3]);
      BX_MIDPOINT (dst[k_dst+2], dst[k_dst+1], v           );
      BX_MIDPOINT (dst[k_dst+4], dst[k_dst+5], v           );
      BX_MIDPOINT (dst[k_dst+3], dst[k_dst+2], dst[k_dst+4]);
      dst [k_dst+6]= src [k_src+3];
    }
    { bx_point * sas= src; src= dst; dst= sas; }
    max_src= k_dst;
  }
  for (k= 0; k <= max_src; k++) {
    x_res [k]= src [k].x;
    y_res [k]= src [k].y;
  }
  return max_src+1; /* = BX_BEZIER_LENGTH */
}


/*-----------------------------------------------------------------------------
 * 
 *-----------------------------------------------------------------------------
 */

int
bx_points_to_bezier2 (int x0, int y0, 
                      int x1, int y1, 
                      int x2, int y2,
                      int x_res [], int y_res [])
{
  bx_point a[BX_BEZIER2_LENGTH], * src= a;
  bx_point b[BX_BEZIER2_LENGTH], * dst= b; 
  int max_src, k_src, k_dst, k, level;

  a[0].x = x0; a[1].x = x1; a[2].x = x2; b[0].x= x0;
  a[0].y = y0; a[1].y = y1; a[2].y = y2; b[0].y= y0;

  for(level= 0, max_src= 2; level < BX_BEZIER2_LEVEL; level++) {
    for (k_src= k_dst= 0; k_src < max_src; k_src+= 2, k_dst+= 4) {
      BX_MIDPOINT (dst[k_dst+1], src[k_src+0], src[k_src+1]);
      BX_MIDPOINT (dst[k_dst+3], src[k_src+1], src[k_src+2]);
      BX_MIDPOINT (dst[k_dst+2], dst[k_dst+1], dst[k_dst+3]);
      dst [k_dst+4]= src [k_src+2];
    }
    { bx_point * sas= src; src= dst; dst= sas; }
    max_src= k_dst;
  }
  for (k= 0; k <= max_src; k++) {
    x_res [k]= src [k].x;
    y_res [k]= src [k].y;
  }
  return max_src+1;/* = BX_BEZIER2_LENGTH */
}

/*-----------------------------------------------------------------------------
 * 
 *-----------------------------------------------------------------------------
 */

void
bx_draw_bezier (bx_window window,
                int x0, int y0, 
                int x1, int y1, 
                int x2, int y2, 
                int x3, int y3)
{
  int x_res [1000], y_res [1000];
  int length= bx_points_to_bezier (x0, y0, 
                                   x1, y1, 
                                   x2, y2, 
                                   x3, y3, 
                                   x_res, y_res);
  bx_draw_polygon (window, length, x_res, y_res, 0, 0);
}

/*-----------------------------------------------------------------------------
 * 
 *-----------------------------------------------------------------------------
 */

void
bx_draw_bezier2 (bx_window window,
                 int x0, int y0, 
                 int x1, int y1, 
                 int x2, int y2)
{
  int x_res [1000], y_res [1000];
  int length= bx_points_to_bezier2 (x0, y0, 
                                    x1, y1, 
                                    x2, y2, 
                                    x_res, y_res);
  bx_draw_polygon (window, length, x_res, y_res, 0, 0);
}

/*-----------------------------------------------------------------------------
 * 
 *              (k)              (k+2)
 *               #                 #
 *              / \               /
 *             /   \             /
 *    q[k-1]  #--#--# p[k]      # q[k+1]  
 *           / t[k]  \         /  
 *          /         \ t[k+1]/
 *  p[k-1] #      q[k] #--#--# p[k+1]
 *        /             \   /  
 *       /               \ /
 *      #                 #
 *    (k-1)             (k+1)
 *
 * third order bezier curve at (t[k],p[k],q[k],t[k+1])
 *-----------------------------------------------------------------------------
 */

int
bx_points_to_curve (int n, int x [], int y [], bx_closure closed, 
                    int x_res[], int y_res[])
{
  bx_point p[BX_MAX_CURVE_POINTS];
  bx_point q[BX_MAX_CURVE_POINTS];
  bx_point t[BX_MAX_CURVE_POINTS];
  int k, j, k_res;
 
  for (k= 0, j= 1; j < n; k++, j++) {
    BX_3RDPOINT_XY (p[k],   x[k],y[k],   x[j],y[j]);
    BX_3RDPOINT_XY (q[k],   x[j],y[j],   x[k],y[k]);
  }
  if (closed) {
    BX_3RDPOINT_XY (p[k],   x[k],y[k],   x[0],y[0]);
    BX_3RDPOINT_XY (q[k],   x[0],y[0],   x[k],y[k]);
  }
  for (k= 1, j=0; k < n-1; k++, j++) { 
    BX_MIDPOINT (t[k],   q[j], p[k]);
  }
  if (closed) {
    BX_MIDPOINT (t[k],   q[j], p[k]);
    BX_MIDPOINT (t[0],   q[k], p[0]);
  } else {
    t[0].x= x[0]; t[k].x= x[k]; 
    t[0].y= y[0]; t[k].y= y[k]; 
  }
  for (k= 0, j= 1, k_res= 0; j < n; k++, j++) {
     k_res+= -1 + bx_points_to_bezier (t[k].x, t[k].y, p[k].x, p[k].y,
                                       q[k].x, q[k].y, t[j].x, t[j].y, 
                                       x_res+k_res, y_res+k_res);
  }
  if (closed) {
    k_res+= -1 + bx_points_to_bezier (t[k].x, t[k].y, p[k].x, p[k].y,
                                      q[k].x, q[k].y, t[0].x, t[0].y, 
                                      x_res+k_res, y_res+k_res);
                                      }
  return k_res+1;
}

/*-----------------------------------------------------------------------------
 * 
 *        (k+1)        (k+3)
 *          #           # 
 *         / \         /
 *   p[k] /   \       /
 *       #     #     #
 *      / p[k+1]\   / p[k+2]
 *     /         \ /
 *    #           #
 *   (k)        (k+2)
 *
 * second order bezier curve at (p[k],(k+1),p[k+1])
 *-----------------------------------------------------------------------------
 */

int
bx_points_to_curve2 (int n, int x [], int y [], bx_closure closed, 
                     int x_res[], int y_res[])
{
  bx_point p[BX_MAX_CURVE2_POINTS];
  int j, k, k_res;
 
  for (k= 0, j=1 ; j < n; k++, j++) {
    BX_MIDPOINT_XY (p[k],   x[k],y[k],   x[j],y[j]);
  }
  if (closed) {
    BX_MIDPOINT_XY (p[k], x[k],y[k], x[0],y[0]);
  } else {
    p[0].x= x[0]; p[n-2].x= x[k]; 
    p[0].y= y[0]; p[n-2].y= y[k];  
  }
  for (k= 0, j= 1, k_res= 0; j < n-1; k++, j++) {
    k_res+= -1 + bx_points_to_bezier2 (p[k].x, p[k].y,
                                       x[j],   y[j]  ,
                                       p[j].x, p[j].y, 
                                       x_res+k_res, y_res+k_res);
  }
  if (closed) {
    k_res+= -1 + bx_points_to_bezier2 (p[k].x,p[k].y,
                                       x[j]  ,y[j]  ,
                                       p[j].x,p[j].y, 
                                       x_res+k_res, y_res+k_res);
    k_res+= -1 + bx_points_to_bezier2 (p[j].x, p[j].y,
                                       x[0]  , y[0]  ,
                                       p[0].x, p[0].y, 
                                       x_res+k_res, y_res+k_res);
    
  }
  return k_res+1;
}

/*-----------------------------------------------------------------------------
 * 
 *-----------------------------------------------------------------------------
 */

static int 
bx_distance (double x1, double y1, double x2, double y2)
{
  double dx= x1 - x2;
  double dy= y1 - y2;
  double d= sqrt (dx*dx + dy*dy); 
  return (d < 1) ? 1 : d;
}

#define BX_BETAPOINTS(S,T,X,Y,DIST,K,J,BETA)    \
{                                               \
  double alpha= DIST[K] / (DIST[K] + DIST[J]);  \
  bx_point q;                                   \
  q.x= p[k].x + alpha*(p[j].x - p[k].x);        \
  q.y= p[k].y + alpha*(p[j].y - p[k].y);        \
  S[j].x= X[j]+BETA*(p[k].x-q.x);               \
  S[j].y= Y[j]+BETA*(p[k].y-q.y);               \
  T[j].x= X[j]+BETA*(p[j].x-q.x);               \
  T[j].y= Y[j]+BETA*(p[j].y-q.y);               \
}

int
bx_points_to_curvei (int n, int x [], int y [], 
                     bx_closure closed, double beta,
                     int x_res[], int y_res[])
{
  bx_point p  [BX_MAX_CURVEI_POINTS];
  bx_point s  [BX_MAX_CURVEI_POINTS];
  bx_point t  [BX_MAX_CURVEI_POINTS];
  double dist [BX_MAX_CURVEI_POINTS];
  int j,k, k_res;
 
  for (k= 0; k < n-1; k++) {
    BX_MIDPOINT_XY (p[k], x[k],y[k], x[k+1],y[k+1]);
    dist[k]= bx_distance (x[k],y[k], x[k+1],y[k+1]);
  
  }
  if (closed) {
    BX_MIDPOINT_XY (p[k], x[k],y[k], x[0],y[0]);
    dist[k]= bx_distance (x[k],y[k], x[0],y[0]);
  }
  for (k= 0, j=1; j < n-1; k++, j++) 
    BX_BETAPOINTS (s, t, x, y, dist, k, j, beta);

  if (closed) {
    k= n-2; j=n-1; BX_BETAPOINTS (s, t, x, y, dist, k, j, beta);
    k= n-1; j=0  ; BX_BETAPOINTS (s, t, x, y, dist, k, j, beta);
  } else {
    BX_3RDPOINT_XY (t[0  ], x[0  ], y[0  ], x[1  ], y[1  ]);
    BX_3RDPOINT_XY (s[n-1], x[n-1], y[n-1], x[n-2], y[n-2]);
  }
  for (k= 0, j= 1, k_res= 0; j < n; k++, j++) {
    k_res+= -1 + bx_points_to_bezier (x[k]  , y[k]  ,
                                      t[k].x, t[k].y,
                                      s[j].x, s[j].y,
                                      x[j]  , y[j]  , 
                                      x_res+k_res, 
                                      y_res+k_res);
  }
  if (closed) {
    k_res+= -1 + bx_points_to_bezier (x[k]  , y[k]  ,
                                      t[k].x, t[k].y,
                                      s[0].x, s[0].y,
                                      x[0]  , y[0]  , 
                                      x_res+k_res, 
                                      y_res+k_res);
  }
  return k_res+1;
}

/*-----------------------------------------------------------------------------
 * 
 *-----------------------------------------------------------------------------
 */

void 
bx_draw_curve (bx_window window, int n, int x [], int y [], 
               bx_closure closed, bx_fill filled)
{
  int x_res [BX_CURVE_LENGTH];
  int y_res [BX_CURVE_LENGTH];
  int length;

  if (n == 0) return;
  if (n == 1) { bx_draw_point (window, x[0], y [0]); return; }
  if (n == 2) { bx_draw_line  (window, x[0], y [0], x[1], y[1]); return; }

  length= bx_points_to_curve (n, x, y, closed, x_res, y_res);
  bx_draw_polygon (window, length, x_res, y_res, 0, filled);
}

/*-----------------------------------------------------------------------------
 * 
 *-----------------------------------------------------------------------------
 */

void 
bx_draw_curve2 (bx_window window, int n, int x [], int y [],
                bx_closure closed, bx_fill filled)
{
  int x_res [BX_CURVE2_LENGTH];
  int y_res [BX_CURVE2_LENGTH];
  int length;
  if (n == 0) return;
  if (n == 1) { bx_draw_point (window, x[0], y [0]); return; }
  if (n == 2) { bx_draw_line  (window, x[0], y [0], x[1], y[1]); return; }

  length= bx_points_to_curve2 (n, x, y, closed, x_res, y_res);
  bx_draw_polygon (window, length, x_res, y_res, 0, filled);

}
/*-----------------------------------------------------------------------------
 * 
 *-----------------------------------------------------------------------------
 */

void 
bx_draw_curvei (bx_window window, int n, int x [], int y [], 
                bx_closure closed, bx_fill filled, double beta)
{
  int x_res [BX_CURVEI_LENGTH];
  int y_res [BX_CURVEI_LENGTH];
  int length;

  if (n == 0) return;
  if (n == 1) { bx_draw_point (window, x[0], y [0]); return; }
  if (n == 2) { bx_draw_line  (window, x[0], y [0], x[1], y[1]); return; }

  length= bx_points_to_curvei (n, x, y, closed, beta, x_res, y_res);
  bx_draw_polygon (window, length, x_res, y_res, 0, filled);
}

/*-----------------------------------------------------------------------------
 * 
 *-----------------------------------------------------------------------------
 */

int
bx_red_depth (void) {
  return  myx.trueColor.red.length;
}
int
bx_green_depth (void) {
  return myx.trueColor.green.length;
}

int
bx_blue_depth (void) {
  return  myx.trueColor.blue.length;
}

int
bx_red_shift (void) {
  return myx.trueColor.red.shift;
}
int
bx_green_shift (void) {
  return myx.trueColor.green.shift;
}

int
bx_blue_shift (void) {
  return myx.trueColor.blue.shift;
}

/*-----------------------------------------------------------------------------
 * 
 *-----------------------------------------------------------------------------
 */

bx_color
bx_rgb_to_color (unsigned char r, unsigned char g, unsigned char b)
{
  bx_color color;
  color.pixel= MyX_RGB (& myx, r, g, b);
  return color;
}

bx_color 
bx_mean_color (bx_color c1, bx_color c2)
{
  bx_color color;
  color.pixel= MyX_AverageColors (& myx, c1.pixel, c2.pixel);
  return color;
}

bx_color 
bx_blend_color   (bx_color c1,  double alpha, bx_color c2)
{
  bx_color color;
  color.pixel= MyX_BlendColors (& myx, c1.pixel, alpha, c2.pixel);
  return color;
}


bx_color bx_white(void) {return bx_rgb_to_color(0xFF,0xFF,0xFF);}
bx_color bx_black(void) {return bx_rgb_to_color(0x00,0x00,0x00);}

bx_color bx_gray     (void) {return bx_rgb_to_color(0x80,0x80,0x80);}
bx_color bx_pale_gray(void) {return bx_rgb_to_color(0xC0,0xC0,0xC0);}
bx_color bx_dark_gray(void) {return bx_rgb_to_color(0x40,0x40,0x40);}

bx_color bx_red     (void) {return bx_rgb_to_color(0xFF,0x00,0x00);}
bx_color bx_pale_red(void) {return bx_rgb_to_color(0xFF,0x80,0x80);}
bx_color bx_dark_red(void) {return bx_rgb_to_color(0x80,0x00,0x00);}

bx_color bx_green     (void) {return bx_rgb_to_color(0x00,0xFF,0x00);}
bx_color bx_pale_green(void) {return bx_rgb_to_color(0x80,0xFF,0x80);}
bx_color bx_dark_green(void) {return bx_rgb_to_color(0x00,0x80,0x00);}

bx_color bx_blue     (void) {return bx_rgb_to_color(0x00,0x00,0xFF);}
bx_color bx_pale_blue(void) {return bx_rgb_to_color(0x80,0x80,0xFF);} 
bx_color bx_dark_blue(void) {return bx_rgb_to_color(0x00,0x00,0x80);}

bx_color bx_yellow     (void) {return bx_rgb_to_color(0xFF,0xFF,0x00);}
bx_color bx_pale_yellow(void) {return bx_rgb_to_color(0xFF,0xFF,0xC0);}
bx_color bx_dark_yellow(void) {return bx_rgb_to_color(0x80,0x80,0x00);}

bx_color bx_cyan     (void) {return bx_rgb_to_color(0x00,0xFF,0xFF);}
bx_color bx_pale_cyan(void) {return bx_rgb_to_color(0xC0,0xFF,0xFF);}
bx_color bx_dark_cyan(void) {return bx_rgb_to_color(0x00,0x80,0x80);}

bx_color bx_magenta     (void) {return bx_rgb_to_color(0xFF,0x00,0xFF);}
bx_color bx_pale_magenta(void) {return bx_rgb_to_color(0xFF,0xC0,0xFF);}
bx_color bx_dark_magenta(void) {return bx_rgb_to_color(0xC0,0x00,0xC0);}

bx_color bx_orange     (void) {return bx_rgb_to_color(0xFF,0x80,0x00);}
bx_color bx_pale_orange(void) {return bx_rgb_to_color(0xFF,0xC0,0x80);}
bx_color bx_dark_orange(void) {return bx_rgb_to_color(0x80,0x40,0x00);}

bx_color bx_pink     (void) {return bx_rgb_to_color(0xFF,0x00,0x80);}
bx_color bx_pale_pink(void) {return bx_rgb_to_color(0xFF,0x80,0xC0);}
bx_color bx_dark_pink(void) {return bx_rgb_to_color(0x80,0x00,0x40);}

bx_color bx_violet     (void) {return bx_rgb_to_color(0x80,0x00,0xFF);}
bx_color bx_pale_violet(void) {return bx_rgb_to_color(0xC0,0x80,0xFF);}
bx_color bx_dark_violet(void) {return bx_rgb_to_color(0x40,0x00,0x80);}

bx_color bx_azure     (void) {return bx_rgb_to_color (0x00, 0x80,0xFF);}
bx_color bx_pale_azure(void) {return bx_rgb_to_color (0x80, 0xC0,0xFF);}
bx_color bx_dark_azure(void) {return bx_rgb_to_color (0x00, 0x40,0x80);}

bx_color bx_spring     (void) {return bx_rgb_to_color (0x00, 0xFF,0x80);}
bx_color bx_pale_spring(void) {return bx_rgb_to_color (0x80, 0xFF,0xC0);}
bx_color bx_dark_spring(void) {return bx_rgb_to_color (0x00, 0x80,0x40);}

bx_color bx_lawn     (void) {return bx_rgb_to_color (0x80, 0xFF,0x00);}
bx_color bx_pale_lawn(void) {return bx_rgb_to_color (0xC0, 0xFF,0x80);}
bx_color bx_dark_lawn(void) {return bx_rgb_to_color (0x40, 0x80,0x00);}



/*-----------------------------------------------------------------------------
 * 
 *-----------------------------------------------------------------------------
 */

static  void
bx_set_color_aux (GC gc, bx_color color) 
{
  XSetForeground (myx.display, gc, color.pixel);
}

void
bx_set_color (bx_color color) 
{
  bx_set_color_aux (myx.gc, color); 
}

/*-----------------------------------------------------------------------------
 * 
 *-----------------------------------------------------------------------------
 */

static  void
bx_set_color2_aux (GC gc, bx_color color) 
{
  XSetBackground (myx.display, gc, color.pixel);
}

 void
bx_set_color2 (bx_color color) 
{
  bx_set_color2_aux (myx.gc, color); 
}

/*-----------------------------------------------------------------------------
 * 
 *-----------------------------------------------------------------------------
 */

static  void
bx_set_rgb_aux (GC gc, unsigned char r, unsigned char g, unsigned char b) 
{
  bx_color color= bx_rgb_to_color (r, g, b);
  bx_set_color_aux (gc, color); 
}

void
bx_set_rgb (unsigned char r, unsigned char g, unsigned char b) 
{
  bx_set_rgb_aux (myx.gc, r, g, b); 
}

/*-----------------------------------------------------------------------------
 * 
 *-----------------------------------------------------------------------------
 */

static void
bx_set_rgb2_aux (GC gc, unsigned char r, unsigned char g, unsigned char b) 
{
  bx_color color= bx_rgb_to_color (r, g, b);
  bx_set_color2_aux (gc, color); 
}

void
bx_set_rgb2 (unsigned char r, unsigned char g, unsigned char b) 
{
  bx_set_rgb2_aux (myx.gc, r, g, b); 
}

/*-----------------------------------------------------------------------------
 *
 *-----------------------------------------------------------------------------
 */

static void 
bx_set_line_style_aux (GC gc, bx_line_style line_style)
{
  XGCValues values;
  values.line_style= line_style;
  XChangeGC (myx.display, gc, GCLineStyle, & values);
}

void 
bx_set_line_style (bx_line_style line_style)
{
  bx_set_line_style_aux (myx.gc, line_style);
}
/*-----------------------------------------------------------------------------
 *
 *-----------------------------------------------------------------------------
 */

static void 
bx_set_fill_rule_aux (GC gc, bx_fill_rule fill_rule)
{
  XSetFillRule (myx.display, gc, fill_rule);
}

void 
bx_set_fill_rule (bx_fill_rule fill_rule)
{
  bx_set_fill_rule_aux (myx.gc, fill_rule);
}
/*-----------------------------------------------------------------------------
 *
 *-----------------------------------------------------------------------------
 */

static void 
bx_set_fill_style_aux (GC gc, bx_fill_style fill_style)
{
  XSetFillStyle (myx.display, gc, !! fill_style);
}

void 
bx_set_fill_style (bx_fill_style fill_style)
{
  bx_set_fill_style_aux (myx.gc, fill_style);
}


/*-----------------------------------------------------------------------------
 *
 *-----------------------------------------------------------------------------
 */

static void  
bx_set_tile_image_aux (GC gc, bx_image image)
{
  XSetTile (myx.display, gc, image.stuff->image);
}

void  
bx_set_tile_image (bx_image image)
{
  bx_set_tile_image_aux (myx.gc, image);
}

/*-----------------------------------------------------------------------------
 *
 *-----------------------------------------------------------------------------
 */

static void  
bx_set_tile_origin_aux (GC gc, int x, int y)
{
  XSetTSOrigin (myx.display, gc, x, y);
}

void  
bx_set_tile_origin (int x, int y)
{
  bx_set_tile_origin_aux (myx.gc, x, y);
}


/*-----------------------------------------------------------------------------
 *
 *-----------------------------------------------------------------------------
 */

static void 
bx_set_line_width_aux (GC gc, unsigned width)
{
  XGCValues values;
  values.line_width= width;
  XChangeGC (myx.display, gc, GCLineWidth, & values);
}

 void 
bx_set_line_width (unsigned width)
{
  bx_set_line_width_aux (myx.gc, width);
}
/*-----------------------------------------------------------------------------
 *
 *-----------------------------------------------------------------------------
 */

static void
bx_set_effect_aux (GC gc, unsigned mode)
{
  XSetFunction (myx.display, gc, mode % 16);
}

void
bx_set_effect (unsigned mode)
{
  bx_set_effect_aux (myx.gc, mode);
}


/*-----------------------------------------------------------------------------
 * dessiner_reglettes
 *-----------------------------------------------------------------------------
 */

static void 
bx_draw_rulers (bx_window win)
{
  int x= win.stuff->x;
  int y= win.stuff->y;
  int xr= x + win.stuff->width;
  int yr= y + win.stuff->height;
 
  unsigned k; 

  bx_set_color_aux (myx.gc_sys, bx_blue()); 

  for (k= 0; k <= win.stuff->width; k+= 10) {
    int mark= (k % 50 == 0);
    if (y-1 < 0) continue; /* prevents bug for lines on some systems */
    bx_draw_line_sys (win, x+k, y -1, x+k, mark ? y -10 : y -5);
    bx_draw_line_sys (win, x+k, yr+1, x+k, mark ? yr+10 : yr+5);
    if (! mark) continue;
    bx_draw_text_sys_f (win, x+k-20, y -15, "%4d", k); 
    bx_draw_text_sys_f (win, x+k-20, yr+25, "%4d", k); 
  }

  for (k= 0; k <= win.stuff->height; k+= 10) {
    int mark= (k % 50 == 0);
    if (x-1 < 0) continue; /* prevents bug for lines on some systems */
    bx_draw_line_sys (win, x -1, y+k, mark ? x -10 : x -5, y+k);
    bx_draw_line_sys (win, xr+1, y+k, mark ? xr+10 : xr+5, y+k);
    if (! mark) continue;
    bx_draw_text_sys_f (win, x -40, y+k+5, "%4d", k); 
    bx_draw_text_sys_f (win, xr+15, y+k+5, "%4d", k); 
  }

  bx_draw_text_sys (win, x + 10, yr + 30 + 10, 
                    "bx-lib version 1.0 (c) Regis Barbanchon 2008");
}

/*-----------------------------------------------------------------------------
 * 
 *-----------------------------------------------------------------------------
 */

static void 
bx_draw_suburb (bx_window win)
{
  int x= win.stuff->x;
  int y= win.stuff->y;
  int xr= x + win.stuff->width;
  int yr= y + win.stuff->height;
 
  if (x > 0) {
    bx_set_color_aux (myx.gc_sys, bx_gray ());
    bx_draw_box_sys (win,  0, 0, x  , win.stuff->win_height, 1);
    bx_draw_box_sys (win, xr, 0, x+1, win.stuff->win_height, 1);
  } 
  if (y > 0) {
    bx_set_color_aux (myx.gc_sys, bx_gray ());
    bx_draw_box_sys (win, x > 0 ? x : 0,  0, win.stuff->width, y  , 1);
    bx_draw_box_sys (win, x > 0 ? x : 0, yr, win.stuff->width, y+1, 1);
  }
}

/*-----------------------------------------------------------------------------
 * 
 *-----------------------------------------------------------------------------
 */
/*
  XCopyArea (myx.display, 
	     win.stuff->front, win.stuff->window, myx.gc_sys,
	     0, 0, win.stuff->width, win.stuff->height, 
             win.stuff->x, win.stuff->y);
*/

void
bx_update_canvas (bx_window win) 
{

  XCopyArea (myx.display, 
	     win.stuff->front, win.stuff->window, myx.gc_sys,/*!!!*/
	     0, 0, win.stuff->width, win.stuff->height, 
             win.stuff->x, win.stuff->y);

  bx_draw_suburb (win); 

  bx_draw_rulers (win);

}

/*-----------------------------------------------------------------------------
 * 
 *-----------------------------------------------------------------------------
 */

void
bx_clear_canvas (bx_window win, bx_color color) 
{
  bx_set_color_aux (myx.gc_sys, color);
  bx_draw_box_aux (win.stuff->back, myx.gc_sys, 
                   0, 0, 
                   bx_canvas_width  (win), 
                   bx_canvas_height (win), 1);
}

/*-----------------------------------------------------------------------------
 * 
 *-----------------------------------------------------------------------------
 */

void 
bx_wait (unsigned long msecs)
{
  struct timeval max_delay; 
  max_delay.tv_sec =  msecs / 1000;
  max_delay.tv_usec= (msecs % 1000) * 1000;
  MyX_WaitEvents (& myx, & max_delay);
}

void
bx_swap_canvas (bx_window win, unsigned long msecs) 
{
  struct timeval max_delay; 
  Pixmap p;
  
  p= win.stuff->back;
  win.stuff->back= win.stuff->front;
  win.stuff->front= p;
 
  bx_update_canvas (win);

  max_delay.tv_sec =  msecs / 1000;
  max_delay.tv_usec= (msecs % 1000) * 1000;

  MyX_WaitEvents (& myx, & max_delay);
 
}

void
bx_show_canvas (bx_window win, unsigned long msecs) 
{
  struct timeval max_delay; 

  XCopyArea (myx.display, 
	     win.stuff->back, win.stuff->front, myx.gc_sys,/*!!!*/
	     0, 0, win.stuff->width, win.stuff->height, 0, 0);

  bx_update_canvas (win);

  max_delay.tv_sec =  msecs / 1000;
  max_delay.tv_usec= (msecs % 1000) * 1000;

  MyX_WaitEvents (& myx, & max_delay);
 
}


/*-----------------------------------------------------------------------------
 *
 *-----------------------------------------------------------------------------
 */

void
bx_free_image (bx_image image)
{
  free (image.stuff);
}

bx_image
bx_load_image (int transp, char const name [])
{
  bx_image image;
  int status;

  image.stuff= malloc (sizeof * image.stuff);
  if (! image.stuff) {
    fprintf (stderr, "bx_load_image() failed: no memory.\n");
    exit (1);
  }
  status= MyX_LoadSprite (& myx, name, image.stuff, transp);
  if (status) {
    fprintf (stderr, "bx_load_image() failed for file %s\n", name);
    free (image.stuff);
    exit (1);
  }
  return image;
}

#define BX_MAX_BUF 2048

bx_image
bx_load_image_f (int transp, char const format [], ...)
{
  char buffer [BX_MAX_BUF];
  va_list args;

  va_start (args, format);
  vsprintf (buffer, format, args);
  va_end (args);
  return bx_load_image (transp, buffer);
}

#undef BX_MAX_BUF

/*-----------------------------------------------------------------------------
 *
 *-----------------------------------------------------------------------------
 */

unsigned
bx_image_width (bx_image image)
{
  return image.stuff->width;
}

unsigned 
bx_image_height (bx_image image)
{
  return image.stuff->height; 
}

/*-----------------------------------------------------------------------------
 *
 *-----------------------------------------------------------------------------
 */

void 
bx_flip_image (bx_image image, int xflip, int yflip)
{
  MyX_XYFlipSprite (& myx, image.stuff, xflip, yflip);
}

void 
bx_draw_image (bx_window win, bx_image image, int x, int y)
{
  MyX_PutSprite (& myx, win.stuff->back, 
                 myx.gc_sys, image.stuff, x, y); 
}

void 
bx_draw_image_fx (bx_window win, bx_image image, int x, int y, bx_effect fx)
{
  MyX_PutFxSprite (& myx, win.stuff->back, 
                   myx.gc_sys, fx, image.stuff, x, y); 
}


/*-----------------------------------------------------------------------------
 * mouse control
 *-----------------------------------------------------------------------------
 */

bx_mouse 
bx_read_mouse (bx_window win) {
  bx_mouse mouse;
 
  XQueryPointer (myx.display, win.stuff->window, 
		 & mouse.root, & mouse.child,
		 & mouse.xAbs, & mouse.yAbs, 
		 & mouse.xRel, & mouse.yRel, 
		 & mouse.keyMask);

  mouse.x = mouse.xRel - win.stuff->x;
  mouse.y = mouse.yRel - win.stuff->y;

  return mouse;
}

int bx_mouse_x       (bx_mouse mouse) { return mouse.x;    }
int bx_mouse_xcanvas (bx_mouse mouse) { return mouse.x;    }
int bx_mouse_xwindow (bx_mouse mouse) { return mouse.xRel; }
int bx_mouse_xscreen (bx_mouse mouse) { return mouse.xAbs; }

int bx_mouse_y       (bx_mouse mouse) { return mouse.y;    }
int bx_mouse_ycanvas (bx_mouse mouse) { return mouse.y;    }
int bx_mouse_ywindow (bx_mouse mouse) { return mouse.yRel; }
int bx_mouse_yscreen (bx_mouse mouse) { return mouse.yAbs; }

int bx_mouse_b1 (bx_mouse m){ return !!(m.keyMask & Button1Mask); }
int bx_mouse_b2 (bx_mouse m){ return !!(m.keyMask & Button2Mask); }
int bx_mouse_b3 (bx_mouse m){ return !!(m.keyMask & Button3Mask); }
int bx_mouse_b4 (bx_mouse m){ return !!(m.keyMask & Button4Mask); }
int bx_mouse_b5 (bx_mouse m){ return !!(m.keyMask & Button5Mask); }

int bx_mouse_b (bx_mouse m, int b)
{
  switch (b) {
  case 1: return bx_mouse_b1 (m); 
  case 2: return bx_mouse_b2 (m); 
  case 3: return bx_mouse_b3 (m); 
  case 4: return bx_mouse_b4 (m); 
  case 5: return bx_mouse_b5 (m); 
  default: return 0;
  }
}

int bx_mouse_shift (bx_mouse m){ return !!(m.keyMask &   ShiftMask); }
int bx_mouse_ctrl  (bx_mouse m){ return !!(m.keyMask & ControlMask); }
int bx_mouse_alt   (bx_mouse m){ return !!(m.keyMask &    Mod1Mask); }
int bx_mouse_altgr (bx_mouse m){ return !!(m.keyMask &    Mod5Mask); }

/*-----------------------------------------------------------------------------
 * keyboard control
 *-----------------------------------------------------------------------------
 */

bx_key 
bx_read_key (void)
{
  bx_key key;
  XQueryKeymap (myx.display, (char *) key.bits);
  return key;
}

char *
bx_keycode_to_keyname (unsigned char keycode)
{
  int keysym= XkbKeycodeToKeysym (myx.display, keycode, 0, 0);
  return XKeysymToString (keysym);
}

unsigned char
bx_keyname_to_keycode (char keyname [])
{
  int keysym= XStringToKeysym (keyname);
  return XKeysymToKeycode (myx.display, keysym);
}

int 
bx_keycode_pressed (bx_key key, unsigned char keycode)
{
  unsigned char num_byte= (keycode & 0xFF) / CHAR_BIT;
  unsigned char num_bit = (keycode & 0xFF) % CHAR_BIT;
  return !! (key.bits [num_byte] & (1 << num_bit));
}

int 
bx_keyname_pressed (bx_key key, char keyname [])
{
  unsigned char keycode= bx_keyname_to_keycode (keyname);
  return bx_keycode_pressed (key, keycode);
}

char *
bx_keynames (bx_key key)
{
  static char names [4096];
  int k, offset= 0;
  for (k= 0; k < 32 * CHAR_BIT; k++) {
    if (bx_keycode_pressed (key, k)) {
      char * keyname=  bx_keycode_to_keyname (k);
      sprintf (names + offset, "%s ", keyname);
      offset += strlen (keyname) + 1;
    }
  }
  return names;
}

int 
bx_key_number (bx_key key) {
  int k, number= 0;
  for (k= 0; k < 32 * CHAR_BIT; k++) {
    number += bx_keycode_pressed (key, k); 
  }
  return number;
}

void 
bx_set_autorepeat (int activate)
{
  MyX_IgnoreAutoRepeat (& myx, ! activate);
}

int 
bx_get_autorepeat (void)
{
  return ! myx.ignoreAutoRepeat;
}


/*-----------------------------------------------------------------------------
 * directories
 *-----------------------------------------------------------------------------
 */

static int
bx_compare_filenames (void const * a,  void const * b)
{
  char * const * x= a;
  char * const * y= b;
  
  return strcmp (* x, * y);
}

bx_dir
bx_open_dir (char const path [])
{
  bx_dir rep;
  DIR * dir;
  struct dirent * dirent;
  long length;

  rep.stuff= malloc (sizeof * rep.stuff);
  if (! rep.stuff) {
    fprintf (stderr, "bx_open_dir() failed: no memory.\n");
    exit (1);
  }
  dir= opendir (path);
  if (! dir) {
    fprintf (stderr, "bx_open_dir() failed for %s.\n", path);
    free (dir); exit (1);
  }
  rep.stuff->nb_chars  = 0;
  rep.stuff->nb_entries= 0;
  for (;;) {
    dirent= readdir (dir);
    if (! dirent) break;

    if (! strcmp (dirent->d_name, "." )) continue;
    if (! strcmp (dirent->d_name, "..")) continue;

    if (rep.stuff->nb_entries >= BX_MAX_DIR_ENTRIES) {
      fprintf (stderr, "bx_open_dir() failed: too many files.\n");
      free (dir); exit (1);
    }
    length= strlen (dirent->d_name);
    if (rep.stuff->nb_chars + length + 1 > BX_MAX_DIR_BUFFER) {
      fprintf (stderr, "bx_open_dir() failed: filenames too long.\n");
      free (dir); exit (1);
    }
    strcpy (rep.stuff->buffer + rep.stuff->nb_chars, dirent->d_name);
    rep.stuff->entries [rep.stuff->nb_entries]= 
      rep.stuff->buffer + rep.stuff->nb_chars;
    rep.stuff->nb_chars += strlen  (dirent->d_name) + 1;
    rep.stuff->nb_entries ++;
  }
  qsort (rep.stuff->entries, rep.stuff->nb_entries, 
	 sizeof * rep.stuff->entries, bx_compare_filenames);
  return rep;
}

#define BX_MAX_BUF 2048

bx_dir
bx_open_dir_f (char const format [], ...)
{
  char buffer [BX_MAX_BUF];
  va_list args;
  va_start (args, format);
  vsprintf (buffer, format, args);
  va_end (args);
  return bx_open_dir_f (buffer);
}

#undef BX_MAX_BUF 

void
bx_close_dir (bx_dir dir)
{
  free (dir.stuff);
}

int
bx_dir_length (bx_dir dir)
{
  return dir.stuff->nb_entries; 
}

char *
bx_dir_filename (bx_dir dir, int index)
{
  return dir.stuff->entries [index];
}

/*-----------------------------------------------------------------------------
 * audio
 *-----------------------------------------------------------------------------
 */

#ifdef BX_AUDIO

struct bx_audio_stuff {
  SDL_AudioSpec file_spec; 
  Uint8 * unconverted_buffer; 
  Uint32 unconverted_length; 
  /* converted sound length is in converter.len_cvt */
  /* converted sound buffer is in converter.buf */
  SDL_AudioCVT converter; 
};

typedef struct Track {
  bx_audio sound;
  int index;    /* position of fragment sent to mixer */
  int die_at;   /* die at this position (index+die_at), or -1: inf loop  */ 
  double volume; /* ratio between 0 and 1 */
  unsigned long start_date;
  int has_started;
} Track;

#define MAX_TRACKS 20
static int nb_tracks;
static Track tracks [MAX_TRACKS];

#define AUDIO_BUFFER_SIZE   1024
static SDL_AudioSpec spec_desired; 
static SDL_AudioSpec spec_obtained;

static void MyAudioFiller     (void *  data, Uint8 * stream, int length);
static void MyAudioFiller_Aux (int num_track, Uint8 * stream, int length);

void
bx_init_audio (void)
{
  spec_desired.freq     = 22050;
  spec_desired.format   = AUDIO_S16;
  spec_desired.channels = 2;
  spec_desired.samples  = AUDIO_BUFFER_SIZE;
  spec_desired.callback = MyAudioFiller;
  spec_desired.userdata = NULL;
  
  if (SDL_OpenAudio (& spec_desired, & spec_obtained) < 0 ) 
  {
    fprintf(stderr, "bx_init_audio() with SDL failed: %s\n", SDL_GetError());
    exit (1);
  }
  nb_tracks= 0;
}

bx_audio
bx_load_audio (char const filename []) 
{
  bx_audio audio;

  audio.stuff= malloc (sizeof * audio.stuff);
  if(! SDL_LoadWAV (filename, 
		    & audio.stuff->file_spec, 
		    & audio.stuff->unconverted_buffer, 
		    & audio.stuff->unconverted_length)) 
  {
    printf("bx_load_audio() failed with file %s: %s\n", 
           filename, SDL_GetError ());
    exit (1);
  }
  SDL_BuildAudioCVT (& audio.stuff->converter, 
                     audio.stuff->file_spec.format, 
                     audio.stuff->file_spec.channels, 
                     audio.stuff->file_spec.freq,
		     spec_obtained.format, 
		     spec_obtained.channels, 
		     spec_obtained.freq);

  audio.stuff->converter.buf = malloc (audio.stuff->unconverted_length * 
                                       audio.stuff->converter.len_mult);

  audio.stuff->converter.len = audio.stuff->unconverted_length; 

  memcpy (audio.stuff->converter.buf, 
          audio.stuff->unconverted_buffer, 
          audio.stuff->unconverted_length);
  SDL_ConvertAudio (& audio.stuff->converter); 

  return audio;
}


#define BX_MAX_BUF 2048

bx_audio
bx_load_audio_f (char const format [], ...)
{
  char buffer [BX_MAX_BUF];
  va_list args;
  va_start (args, format);
  vsprintf (buffer, format, args);
  va_end (args);
  return bx_load_audio (buffer);
}

#undef BX_MAX_BUF

void
bx_play_audio (bx_audio audio, double volume,  
               unsigned long millisec, int ad_infinitum) 
{
  struct timeval now;

  if (nb_tracks >= MAX_TRACKS) 
    return;
  tracks[nb_tracks].sound= audio;
  tracks[nb_tracks].index= 0;
  tracks[nb_tracks].volume= volume;
  tracks[nb_tracks].die_at= ad_infinitum ? -1 : audio.stuff->converter.len_cvt;

  gettimeofday (& now, NULL);
  /* danger ? */
  tracks[nb_tracks].start_date= now.tv_sec * 1000UL + now.tv_usec / 1000UL;
  tracks[nb_tracks].start_date += millisec;
  tracks[nb_tracks].has_started= 0;
  nb_tracks ++;
  SDL_PauseAudio(0);
}

static void 
MyAudioFiller (void * data UNUSED, Uint8 * stream, int length)
{
  struct timeval now;
  ULong date;
  int k;
  gettimeofday (& now, NULL); 
  /* danger ? */
  date= now.tv_sec  * 1000UL + now.tv_usec / 1000UL;
  memset (stream, 0, length); 
  for (k= 0; k < nb_tracks; k++) {
    if (tracks [k].start_date <= date) {
      tracks [k].has_started= 1;
      MyAudioFiller_Aux (k, stream, length);
    }
  }
}

void 
MyAudioFiller_Aux (int k, Uint8 * stream, int length)
{
  /* play less than length if near the end of track */
  if (tracks[k].die_at != -1 && 
      length >= tracks[k].die_at) {
    length= tracks[k].die_at;
  }
  if (tracks[k].index + length >= 
      tracks[k].sound.stuff->converter.len_cvt) {
    length = tracks[k].sound.stuff->converter.len_cvt - tracks [k].index;
  }
  /* send to mixer */
  SDL_MixAudio (stream, 
                tracks[k].sound.stuff->converter.buf + 
                tracks[k].index, 
                length, 
                tracks[k].volume * SDL_MIX_MAXVOLUME);

  /* update index and die_at */
  tracks[k].index += length;
  if (tracks[k].die_at != -1) { 
    tracks[k].die_at -= length; 
  }

  if (tracks[k].index < tracks[k].sound.stuff->converter.len_cvt && 
      tracks[k].die_at != 0) {
    /* play the rest of the track next time */
    return;
  }

  /* end of track reached or die point reached */
  
  if (tracks[k].die_at == -1) {
    /* rewind to play ad infinitum */
    tracks[k].index=  0;     
    return;
  }
  
  /* kill track */
  if (k != nb_tracks - 1) {  
    tracks [k]= tracks [nb_tracks - 1];
  }
  nb_tracks--;
}

#endif

#ifdef MAIN
int main(void) {
  bx_window window;
  bx_init ();
  window= bx_create_window ("bx test", 10, 10, 100, 100);
  bx_loop ();
  return 0;
}
#endif
