
#include "myx.h"


#ifdef _FILE_OFFSET_BITS
#  undef _FILE_OFFSET_BITS
#endif
#include <wand/magick-wand.h>

const char * MyX_EventString []= {
  "DummyEvent",  
  "DummyEvent", 
  "KeyPress", 
  "KeyRelease", 
  "ButtonPress", 
  "ButtonRelease",
  "MotionNotify",            
  "EnterNotify", 
  "LeaveNotify",            
  "FocusIn",     
  "FocusOut",                
  "KeymapNotify",            
  "Expose",                  
  "GraphicsExpose",          
  "NoExpose",                
  "VisibilityNotify",        
  "CreateNotify",            
  "DestroyNotify",           
  "UnmapNotify",             
  "MapNotify",               
  "MapRequest",              
  "ReparentNotify",          
  "ConfigureNotify",         
  "ConfigureRequest",        
  "GravityNotify",           
  "ResizeRequest",          
  "CirculateNotify",         
  "CirculateRequest",        
  "PropertyNotify",          
  "SelectionClear",          
  "SelectionRequest",        
  "SelectionNotify",         
  "ColormapNotify",          
  "ClientMessage",           
  "MappingNotify",           
  "LASTEvent"     
};                

int
MyX_Init (MyX * myx, char const displayName [])
{
  myx->display= XOpenDisplay (displayName);
  if (! myx->display) return -1;

  myx->screen= XDefaultScreen     (myx->display);
  myx->root  = XDefaultRootWindow (myx->display);
  myx->gc    = XDefaultGC         (myx->display, myx->screen);
  myx->depth = XDefaultDepth      (myx->display, myx->screen);
    
  MyX_InitColor (myx);

  XSetForeground (myx->display, myx->gc, MyX_RGB (myx, 0x00, 0x00, 0x00));
  XSetBackground (myx->display, myx->gc, MyX_RGB (myx, 0xFF, 0xFF, 0xFF));
 
  myx->gc_sys= XCreateGC (myx->display, myx->root, 0x0UL, NULL);
  XCopyGC (myx->display, myx->gc,  ~ 0x0UL, myx->gc_sys);

  XSetGraphicsExposures(myx->display, myx->gc, False);
  XSetGraphicsExposures(myx->display, myx->gc_sys, False);

  myx->font_fixed    = XLoadQueryFont (myx->display, "fixed");
  myx->font_variable = XLoadQueryFont (myx->display, "variable");
  if (! myx->font_fixed   ) fprintf (stderr, "Warning: no fixed font!\n");
  if (! myx->font_variable) fprintf (stderr, "Warning: no variable font!!\n");

  myx->handlersContext= XUniqueContext (); 
  myx->dataContext    = XUniqueContext (); 
  myx->timerQueue.nb_timers= 0;            
 
  myx->do_loop= 1;

  MyX_IgnoreAutoRepeat (myx, 0); 

  /* MagickWandGenesis(); */

  return 0;
}

void
MyX_Clean (MyX * myx) 
{
  XFreeGC (myx->display, myx->gc_sys);
  if (myx->visual->class == PseudoColor)
    XFreeColormap (myx->display, myx->pseudoColor.colormap);
  XCloseDisplay (myx->display);
}

INLINE double
MyRound (double x) 
{
  double c= ceil  (x);
  double f= floor (x);
  return (c-x < x-f) ? c : f;
}  

/*-----------------------------------------------------------------------------
 * fonctions auxilliaires pour MyX_InitColor() en mode PseudoColor...
 * construction d'un cube RGB versatile 8x8x4 dans une Colormap... 
 *-----------------------------------------------------------------------------
 */

static void
MyX_SetPalette (MyX * myx, ULong pixel, int r, int g, int b)
{
  XColor c; 
  c.pixel= pixel;
  c.red  = (int) MyRound (r * 0xFFFF/(double) 7); 
  c.green= (int) MyRound (g * 0xFFFF/(double) 7);  
  c.blue = (int) MyRound (b * 0xFFFF/(double) 3); 
  c.flags= DoRed | DoGreen | DoBlue;
  XStoreColor (myx->display, myx->pseudoColor.colormap, & c);
  myx->pseudoColor.palette [r][g][b]= c.pixel;
  myx->pseudoColor.samples [c.pixel]=  c;
}

static void
MyX_InitPseudoColor (MyX * myx)
{  
  int i, j, k;
  ULong pixel;

  myx->pseudoColor.colormap= 
    XCreateColormap (myx->display, myx->root, myx->visual, AllocAll);  

  for (i= 0, pixel= 0; i < 8; i++)
    for (j= 0; j < 8; j++) 
      for (k= 0; k < 4; k++, pixel++) 
        MyX_SetPalette (myx, pixel, i, j, k); 
  
  myx->trueColor.blue. shift=0;
  myx->trueColor.green.shift=2;
  myx->trueColor.red.  shift=5;

  myx->trueColor.blue. length=2;
  myx->trueColor.green.length=3;
  myx->trueColor.red.  length=3;

  myx->trueColor.blue. mask=0x3 << 0;
  myx->trueColor.green.mask=0x7 << 2;
  myx->trueColor.red.  mask=0x7 << 5;

  myx->trueColor.blue. max=3;
  myx->trueColor.green.max=7;
  myx->trueColor.red.  max=7;
}

/*-----------------------------------------------------------------------------
 * fonctions auxilliaires pour MyX_InitColor() en mode TrueColor...
 * calcul des positions des canaux R,G,B d'un pixel...
 *-----------------------------------------------------------------------------
 */

static void
MyX_InitChannel (MyX_Channel * channel, ULong mask)
{
  channel->mask= mask;
  channel->shift= channel->length= channel->max= 0;  
  if (! mask) return;
  while ( (mask  & 1) == 0 ) { channel->shift  ++; mask >>= 1; }
  while ( (mask  & 1) == 1 ) { channel->length ++; mask >>= 1; }
  channel->max= channel->mask >> channel->shift;
}

static void
MyX_InitTrueColor (MyX * myx) 
{
  MyX_InitChannel (& myx->trueColor.blue , myx->visual-> blue_mask);
  MyX_InitChannel (& myx->trueColor.green, myx->visual->green_mask);
  MyX_InitChannel (& myx->trueColor.red  , myx->visual->  red_mask);
}

/*=============================================================================
 * int MyX_InitColor (MyX * myx) 
 * ============================================================================
 * dtecte la classe visuelle et
 * initialise le mode PseudoColor ou TrueColor
 * retourne un statut d'erreur 0/-1 
 *=============================================================================
 */

int
MyX_InitColor (MyX * myx) 
{
  myx->visual= DefaultVisual (myx->display, myx->screen);
  switch (myx->visual->class) {
  case PseudoColor:  MyX_InitPseudoColor (myx); return  0;
  case   TrueColor:  MyX_InitTrueColor   (myx); return  0;
  default         :  /* visual unsupported */   return -1;
  }
}

/*-----------------------------------------------------------------------------
 * fonctions auxilliaires pour MyX_RGB() et MyX_RGBF en mode PseudoColor...
 *-----------------------------------------------------------------------------
 */

static INLINE ULong
MyX_PseudoColorRGB (MyX_PseudoColor * pc, UChar r, UChar g, UChar b)
{
  return pc->palette [r / 32][g / 32][b / 64];
}

static INLINE ULong
MyX_PseudoColorRGBF (MyX_PseudoColor * pc, double r, double g, double b) 
{ 
  r=  (r <= 0) ? 0 : (r >= 1) ? 1 : r; 
  g=  (g <= 0) ? 0 : (g >= 1) ? 1 : g; 
  b=  (b <= 0) ? 0 : (b >= 1) ? 1 : b; 

  return pc->palette 
    [(int) MyRound (r * 7)]
    [(int) MyRound (g * 7)]  
    [(int) MyRound (b * 3)];
}

/*-----------------------------------------------------------------------------
 * fonctions auxilliaires pour MyX_RGB() et MyX_RGBF en mode TrueColor...
 *-----------------------------------------------------------------------------
 */

static INLINE ULong
MyX_ChannelRatioToMask (MyX_Channel * channel, double ratio)
{
  ULong value= (ULong) MyRound (ratio * channel->max);
/*  if (value > channel->max) value= channel->max; */
  return value << channel->shift;
}

static INLINE ULong
MyX_TrueColorRGBF (MyX_TrueColor * tc, double r, double g, double b) 
{
  r=  (r <= 0) ? 0 : (r >= 1) ? 1 : r; 
  g=  (g <= 0) ? 0 : (g >= 1) ? 1 : g; 
  b=  (b <= 0) ? 0 : (b >= 1) ? 1 : b; 

  return MyX_ChannelRatioToMask (& tc->red  , r)
    |    MyX_ChannelRatioToMask (& tc->green, g)
    |    MyX_ChannelRatioToMask (& tc->blue , b);
}

static INLINE ULong
MyX_TrueColorRGB (MyX_TrueColor * tc, 
                  UChar r, UChar g, UChar b)
{
  return 
    r >> (8 - tc->red  .length)<< tc->red  .shift |
    g >> (8 - tc->green.length)<< tc->green.shift |
    b >> (8 - tc->blue .length)<< tc->blue .shift ;
}

/*=============================================================================
 * ULong MyX_RGB  (MyX * myx, UChar  r, UChar  g, UChar  b)
 * ULong MyX_RGBF (MyX * myx, double r, double g, double b)
 * ============================================================================
 * Obtention d'un pixel  partir de (r,g,b)
 * avec r,g,b entiers entre 0 et 0xFF
 * valable en PseudoColor et TrueColor
 * Nota: MyX_RGB tronque, MyX_RGBF arrondit...
 *=============================================================================
 */

INLINE ULong
MyX_RGB (MyX * myx, UChar r, UChar g, UChar b)
{
  return (myx->visual->class == PseudoColor) ?
    MyX_PseudoColorRGB (& myx->pseudoColor, r, g, b) :
    MyX_TrueColorRGB   (& myx->trueColor  , r, g, b);
}

INLINE ULong
MyX_RGBF (MyX * myx, double r, double g, double b)
{
  return (myx->visual->class == PseudoColor) ?
    MyX_PseudoColorRGBF (& myx->pseudoColor, r, g, b) :
    MyX_TrueColorRGBF   (& myx->trueColor  , r, g, b);
}

/*-----------------------------------------------------------------------------
 * fonction auxilliaire pour MyX_AverageColors()  en mode PseudoColor...
 *-----------------------------------------------------------------------------
 */

INLINE ULong
MyX_PseudoColorAverage (MyX_PseudoColor * pc, ULong p1, ULong p2)
{
  ULong r1= pc->samples [p1].red; 
  ULong r2= pc->samples [p2].red; 

  ULong g1= pc->samples [p1].green; 
  ULong g2= pc->samples [p2].green; 
  
  ULong b1= pc->samples [p1].blue; 
  ULong b2= pc->samples [p2].blue; 

  return pc->palette
    [7 * (r1 + r2) >> (8+1)]
    [7 * (g1 + g2) >> (8+1)]
    [3 * (b1 + b2) >> (8+1)];
}



/*-----------------------------------------------------------------------------
 * fonctions auxilliaires pour MyX_AverageColors()  en mode TrueColor...
 *-----------------------------------------------------------------------------
 */

static INLINE ULong
MyX_Channel_PixelComponent (MyX_Channel * channel, ULong pixel)
{
  return ((pixel & channel->mask) >> channel->shift);
}

INLINE ULong 
MyX_TrueColorAverage (MyX_TrueColor * tc, ULong p1, ULong p2)
{
  ULong r1= MyX_Channel_PixelComponent (& tc->red,   p1);
  ULong r2= MyX_Channel_PixelComponent (& tc->red,   p2);

  ULong g1= MyX_Channel_PixelComponent (& tc->green, p1);
  ULong g2= MyX_Channel_PixelComponent (& tc->green, p2);

  ULong b1= MyX_Channel_PixelComponent (& tc->blue,  p1);
  ULong b2= MyX_Channel_PixelComponent (& tc->blue,  p2);
 
  return (r1 + r2)/2 << tc->red.  shift
    |    (g1 + g2)/2 << tc->green.shift
    |    (b1 + b2)/2 << tc->blue. shift; 
}

/*=============================================================================
 * ULong MyX_AverageColors (MyX * myx, ULong p1, ULong p2)
 *=============================================================================
 * renvoie le pixel situ  mi-distance RGB de 2 pixels...
 *=============================================================================
 */

INLINE ULong 
MyX_AverageColors (MyX * myx, ULong p1, ULong p2)
{
  return myx->visual->class == PseudoColor ? 
    MyX_PseudoColorAverage (& myx->pseudoColor, p1, p2) : 
    MyX_TrueColorAverage       (& myx->trueColor  , p1, p2) ; 
}

/*-----------------------------------------------------------------------------
 * fonctions auxiliairex pour MyX_BlendColors() en mode TrueColor...
 *-----------------------------------------------------------------------------
 */

static INLINE double
MyX_PixelToChannelRatio (MyX_Channel * channel, ULong pixel)
{
  return ((pixel & channel->mask) >> channel->shift) 
    /    (double) channel->max;
}

static INLINE void
MyX_TrueColorPixelToRatios (MyX_TrueColor * tc, ULong pixel,
                            double * r, double * g, double * b)
{
  * r= MyX_PixelToChannelRatio (& tc->red,   pixel);
  * g= MyX_PixelToChannelRatio (& tc->green, pixel);
  * b= MyX_PixelToChannelRatio (& tc->blue,  pixel);
}

static INLINE ULong 
MyX_TrueColorBlend (MyX_TrueColor * tc, 
                    ULong p1, double alpha, ULong p2)
{
  double r1,g1,b1;
  double r2,b2,g2;
  MyX_TrueColorPixelToRatios (tc, p1, & r1, & g1, & b1);
  MyX_TrueColorPixelToRatios (tc, p2, & r2, & g2, & b2);
  return  MyX_TrueColorRGBF (tc, 
                             r1 + (r2 - r1) * alpha,
                             g1 + (g2 - g1) * alpha,
                             b1 + (b2 - b1) * alpha);
}

/*-----------------------------------------------------------------------------
 * fonctions auxiliaires pour MyX_BlendColors() en mode PseudoColor... 
 *-----------------------------------------------------------------------------
 */

static INLINE void
MyX_PseudoColorPixelToRatios (MyX_PseudoColor * pc, ULong pixel,
                              double * r, double * g, double * b)
{
  XColor * c= & pc->samples [pixel];
  * r= c->red   / (double) 0xFFFF;
  * g= c->green / (double) 0xFFFF;
  * b= c->blue  / (double) 0xFFFF;
}

static INLINE ULong 
MyX_PseudoColorBlend (MyX_PseudoColor * pc, 
                          ULong p1, double alpha, ULong p2)
{
  double r1,g1,b1;
  double r2,b2,g2;
  MyX_PseudoColorPixelToRatios (pc, p1, & r1, & g1, & b1);
  MyX_PseudoColorPixelToRatios (pc, p2, & r2, & g2, & b2);
  return  MyX_PseudoColorRGBF (pc, 
                               r1 + (r2 - r1) * alpha,
                               g1 + (g2 - g1) * alpha,
                               b1 + (b2 - b1) * alpha);
}

/*=============================================================================
 * ULong MyX_BlendColors (MyX * myx, ULong p1, ULong p2)
 *=============================================================================
 * renvoie le blend de 2 pixels: p= (1.0-alpha)*p1 + alpha*p2
 *=============================================================================
 */

INLINE ULong 
MyX_BlendColors (MyX * myx, ULong p1, double alpha, ULong p2)
{
  return myx->visual->class == PseudoColor ?
    MyX_PseudoColorBlend (& myx->pseudoColor, p1, alpha, p2) :
    MyX_TrueColorBlend   (& myx->trueColor,   p1, alpha, p2);
}


void
MyX_InitHandlers (MyX_Handlers * handlers)
{
  int k;
  for (k= 0; k < MYX_MAX_EVENTS; k++) 
    handlers->table [k]= (MyX_Handler) NULL;
}

void
MyX_SetHandler (MyX_Handlers * handlers, 
                int eventType, MyX_Handler handler)
{
  handlers->table [eventType]= handler;
}

Window
MyX_CreateWindow (MyX * myx, MyX_Handlers * handlers, void * data,
                  Window parent, int x, int y, UInt w, UInt h)
{
  Window window;
  XSizeHints hints;

  window= XCreateSimpleWindow (myx->display, parent, x, y, w, h, 0, 
                               MyX_RGB (myx, 0x00,0x00,0x00), 
                               MyX_RGB (myx, 0xFF,0xFF,0xFF)); 

  hints.flags= USPosition|USSize|PPosition|PSize;
  hints.width = w; hints.x= x;
  hints.height= h; hints.y= y;
  XSetWMNormalHints(myx->display, window, & hints);

  MyX_SetWindowHandlers (myx, window, handlers);
  MyX_SetWindowData     (myx, window, data);

  if (myx->visual->class == PseudoColor)
    XSetWindowColormap (myx->display, window, myx->pseudoColor.colormap);
  return window;
}

void
MyX_DestroyWindowContext (MyX * myx, Window window) 
{
  XDeleteContext (myx->display, window, myx->handlersContext);
  XDeleteContext (myx->display, window, myx->dataContext);
}

void
MyX_SetWindowHandlers (MyX * myx, Window window, MyX_Handlers * handlers)
{
  XSaveContext (myx->display, window, 
                myx->handlersContext, (XPointer) handlers);
}
 
MyX_Handlers *
MyX_GetWindowHandlers (MyX * myx, Window window)
{
  XPointer handlers;
  XFindContext (myx->display, window, 
                myx->handlersContext, & handlers);
  return (MyX_Handlers *) handlers;
}

void
MyX_SetWindowData (MyX * myx, Window window, void * data)
{
  XSaveContext (myx->display, window, 
                myx->dataContext, (XPointer) data); 
}

void * 
MyX_GetWindowData (MyX * myx, Window window)
{
  XPointer data;
  XFindContext (myx->display, window, 
                myx->dataContext, & data);
  return data;
}

void
MyX_QueryTree (MyX * myx, Window w, MyX_Tree * t)
{
  XQueryTree (myx->display, w, 
              & t->root, & t->parent, 
              & t->children, & t->nb_children);
}

void
MyX_GetGeometry (MyX * myx, Drawable d, MyX_Geometry * g)
{
  XGetGeometry (myx->display, d, & g->root,
                & g->x, & g->y, & g->w, & g->h, 
                & g->border, & g->depth);
}

/*=============================================================================
 * void MyX_IgnoreAutoRepeat (MyX * myx, int ignore)
 *=============================================================================
 * permet d'ignorer (ignore=1) ou non (ignore=0)
 * les fake KeyRelease/KeyPress gnrs quand une touche reste enfonce...
 * Ces fakes se suivent dans la file et ont toujours la meme date.
 * Lorsqu'ils sont ignors, les handlers ne sont pas appels.
 *
 * KeyPressMask et KeyRelaseMask doivent etre tous 2 activs
 *=============================================================================
 */

void 
MyX_IgnoreAutoRepeat (MyX * myx, int ignore)
{
  myx->ignoreAutoRepeat= ignore;
}

/*-----------------------------------------------------------------------------
 * un evenement X11 venant d'etre lu est-il un fake KeyRelease ?
 *-----------------------------------------------------------------------------
 */

static int
MyX_Event_IsAutoRepeat (MyX * myx, XEvent * event) 
{
  XEvent next;
  if (event->xany.type != KeyRelease) return 0; 
  if (! XPending (myx->display)) return 0; 
  XPeekEvent (myx->display, & next);
  if (next.xany.type != KeyPress) return 0;
  if (next.xkey.time != event->xkey.time) return 0;
  return 1;}

/*-----------------------------------------------------------------------------
 * dispatche un venement X11 vers son handler s'il y en a un...
 *-----------------------------------------------------------------------------
 */

static void
MyX_Event_Dispatch (MyX * myx, XEvent * event) 
{
  Window window= event->xany.window;
  if (window != None) {
    MyX_Handlers * handlers= MyX_GetWindowHandlers (myx, window);
    if (handlers) {
      MyX_Handler handler=  handlers->table [event->xany.type]; 
      if (handler) {  
        void * data= MyX_GetWindowData (myx, window);
        handler (myx, event, data);
      }
    }
    if (event->xany.type == DestroyNotify) {
      MyX_DestroyWindowContext (myx, window);
    }
  }
}

/*-----------------------------------------------------------------------------
 * dispatche un venement X11 vers son handler s'il y en a un...
 * sauf s'il doit etre ignor:
 * - car c'est un fake KeyRelease et le mode ignoreAutoRepeat est actif
 *-----------------------------------------------------------------------------
 */

static void
MyX_Event_FilterOrDispatch (MyX * myx, XEvent * event) 
{
  if (myx->ignoreAutoRepeat && MyX_Event_IsAutoRepeat (myx, event)) {
    XEvent next;
    XNextEvent (myx->display, & next);
    return;
  } 
  /*
   * if (myx->ic && XFilterEvent (event, None)) return;
   */
  MyX_Event_Dispatch (myx, event);
}

/*-----------------------------------------------------------------------------
 * dispatche un evenement timer vers son handler s'il y en a un...
 *-----------------------------------------------------------------------------
 */

static void
MyX_Timer_Dispatch (MyX * myx, MyX_Timer * timer)
{
 if (timer->handler) {
    timer->handler (myx, timer->data, & timer->expected);
 }
}  

/*-----------------------------------------------------------------------------
 * vrifie si un evenement timer devrait etre trait maintenant...
 *-----------------------------------------------------------------------------
 */

static int
MyX_TimerQueue_IsReady (MyX_TimerQueue * queue, struct timeval * now)
{
  return queue->nb_timers &&
    timercmp (& queue->timers [0].expected, now, <);
}

/*-----------------------------------------------------------------------------
 * extrait le 1er evenement timer de la queue...
 *-----------------------------------------------------------------------------
 */

static void
MyX_TimerQueue_Remove (MyX_TimerQueue * queue, MyX_Timer * timer)
{
  * timer= queue->timers [0];
  queue->nb_timers --;
  memmove (queue->timers, queue->timers + 1, 
           queue->nb_timers * sizeof * timer);
}

/*-----------------------------------------------------------------------------
 * insre un evenement timer dans la queue...
 * celle-ci reste trie sur les dates d'expiration...
 *-----------------------------------------------------------------------------
 */

static int
MyX_TimerQueue_Add (MyX_TimerQueue * queue, MyX_Timer * timer)
{
  ULong k;
  if (queue->nb_timers >= MYX_MAX_TIMERS) return -1;
  for (k= 0; k < queue->nb_timers; k++)
    if (timercmp (& queue->timers [k].expected, & timer->expected, > ))
      break;
  memmove (queue->timers + k + 1, queue->timers + k, 
           (queue->nb_timers-k)* sizeof * timer);
  queue->timers [k]= * timer;
  queue->nb_timers ++;
  return 0;
}

/*=============================================================================
 * int MyX_CallWithinDelayFrom (MyX * myx, MyX_TimerHandler h, void * data,
 *                              struct timeval * delay, struct timeval * from)
 * ============================================================================
 * insre un evenement timer dans la queue de telle sorte que le handler h
 * soit invoqu dans 'delay' sec/usec  partir de l'instant 'from':
 * - si 'from' est NULL : l'instant de reference est 'maintenant'.
 * - si 'delay' est NULL: le dcalage  partir de 'from' est 0 sec/usec
 * - 'data' est transmis  h, ainsi que la date espre d'invocation,
 * ce qui permet de grer/corriger les retards d'invocation.
 *=============================================================================
 */

int
MyX_CallWithinDelay (MyX * myx, MyX_TimerHandler handler, void * data,
                     struct timeval * delay, struct timeval * from)
{
  MyX_Timer timer;
 
  if (from) timer.expected= * from;
  else gettimeofday (& timer.expected, NULL);

  if (delay) timeradd (& timer.expected, delay, & timer.expected);

  timer.handler= handler;
  timer.data= data;
  return MyX_TimerQueue_Add (& myx->timerQueue, & timer);
}


/*-----------------------------------------------------------------------------
 * select() sur le fd du display avec time-out (non NULL)
 * ou 1er timer de la queue 
 *-----------------------------------------------------------------------------
 */

static int
MyX_Select (MyX * myx, struct timeval * max_delay)
{
  int displayFd = ConnectionNumber (myx->display);
  fd_set readFds;
  struct timeval now, delay, expected;

  gettimeofday (& now, NULL); 
  if (myx->timerQueue.nb_timers) {
    expected= myx->timerQueue.timers [0].expected;
    if (timercmp (& expected, & now, <)) timerclear (& delay);
    else timersub (& expected, & now, & delay);
    if (timercmp (max_delay, & delay, <)) delay= * max_delay;
  } else {
    delay= * max_delay;
  }
  FD_ZERO (& readFds); 
  FD_SET (displayFd, & readFds);
  return select (displayFd+1, & readFds, NULL, NULL, & delay);
}

/*-----------------------------------------------------------------------------
 *
 *-----------------------------------------------------------------------------
 */

static void
MyX_ProcessPendingTimers (MyX * myx)
{
  MyX_Timer timer;
  struct timeval now;

  gettimeofday (& now, NULL);
  while (MyX_TimerQueue_IsReady (& myx->timerQueue, & now)) {
    MyX_TimerQueue_Remove (& myx->timerQueue, & timer);
    MyX_Timer_Dispatch (myx, & timer);
  } 
}

static void
MyX_ProcessPendingEvents (MyX * myx)
{
  int k, nb_pending;
  XEvent event;
  do {
    nb_pending= XPending (myx->display);
    for (k= 0; k < nb_pending; k++) {
      XNextEvent (myx->display, & event); 
      MyX_Event_FilterOrDispatch (myx, & event);
    }
  } while (nb_pending);
}

static int
MyX_ProcessSelectableEvents (MyX * myx, struct timeval * bound)
{
 
  struct timeval now, max_delay;
  int res;



  gettimeofday (& now, NULL);
  if (timercmp (& now, bound, >)) return 0;
  timersub (bound , & now, & max_delay);

  res= MyX_Select (myx, & max_delay);
  if      (res == -1) { fprintf (stderr, "select() error...\n"); exit (1); }
  else if (res >   0) { MyX_ProcessPendingEvents (myx); }
  else if (res ==  0) { MyX_ProcessPendingTimers (myx); }

  return 1;
}


void
MyX_WaitEvents (MyX * myx, struct timeval * max_delay) 
{
  struct timeval now, bound;

  gettimeofday (& now, NULL);
  timeradd (& now, max_delay, & bound);

  MyX_ProcessPendingTimers (myx);
  MyX_ProcessPendingEvents (myx);
  while (MyX_ProcessSelectableEvents (myx, & bound)) 
    /* do naught */;
}

/*=============================================================================
 * void MyX_EventLoop (MyX * myx) 
 * boucle sur attente d'evenements X11 ou timer et les traite ou les filtre
 *=============================================================================
 */

void
MyX_EventLoop (MyX * myx) 
{
  struct timeval max_delay;
  max_delay.tv_sec = 5;
  max_delay.tv_usec= 0;
  while (myx->do_loop) {
    MyX_WaitEvents (myx, & max_delay);
  }
}


/*=============================================================================
 *
 *=============================================================================
 */

int
MyX_LoadImage (MyX * myx, char const filename [], 
               XImage ** img, XImage ** msk, UInt * w, UInt * h)
{
#define MAX_ROW 10000
  UChar row [4 * MAX_ROW];
  MagickWand * wand;

  char * img_data, * msk_data, * mode [2]= { "RGB", "RGBA" };
  int width, height, alpha;
  int i, j, k, j_dst, j_src;
  int quantum, bit_quantum;
  ULong img_pix, msk_pix;
  int endian= XImageByteOrder (myx->display);
 
  if (img) * img= NULL;
  if (msk) * msk= NULL;

  if ((wand= NewMagickWand ()) == NULL) return 1;
  if (MagickReadImage(wand, filename) == MagickFalse) return 2;

  alpha= !!msk; /* MagickGetImageMatte  (wand); */

  height= MagickGetImageHeight (wand);
  width=  MagickGetImageWidth  (wand);
  if (width > MAX_ROW) return 3;
  
  quantum= (myx->depth / 8);
  if (quantum == 3) quantum ++;
  bit_quantum= 8 * quantum;
  
  img_data= img ? malloc (height * width * quantum) : NULL;
  msk_data= msk ? malloc (height * width * quantum) : NULL;
  if ((img && !img_data) || (msk && !msk_data)) {
    free (img_data); free (msk_data);
    return 4;
  }

  for (i= 0, j_dst= 0; i < height; i++) {
    MagickExportImagePixels (wand, 0,i, width,1, mode[alpha], CharPixel, row);
    for (j= 0, j_src= 0; j < width; j++, j_src += 3+alpha) {
      if (! alpha || row [j_src+3]) {
	img_pix= MyX_RGB (myx, row [j_src+0], row [j_src+1], row [j_src+2]);
	msk_pix= 0;  /* black contour */
      } else { 
	img_pix=  0; /* black shape */
	msk_pix= ~0; /* white contour */
      }
      if (endian == LSBFirst) {
        for (k= 0; k < bit_quantum; k+= 8, j_dst++) {
          if (img_data) img_data [j_dst]= (img_pix >> k) & 0xFF;
          if (msk_data) msk_data [j_dst]= (msk_pix >> k) & 0xFF;
        }
      } else {
        for (k=  bit_quantum-8; k >= 0; k-= 8, j_dst++) {
          if (img_data) img_data [j_dst]= (img_pix >> k) & 0xFF;
          if (msk_data) msk_data [j_dst]= (msk_pix >> k) & 0xFF;
        }
      }
    }
  }
  DestroyMagickWand (wand);

  if (img) * img= XCreateImage (myx->display, myx->visual, myx->depth, 
                                ZPixmap, 0, img_data, width, height, 
                                bit_quantum, width * quantum);

  if (msk) * msk= XCreateImage (myx->display, myx->visual, myx->depth,
                                ZPixmap, 0, msk_data, width, height, 
                                bit_quantum, width * quantum);

  * w= width; * h= height;
  return 0;
#undef MAX_ROW
}

/*-----------------------------------------------------------------------------
 * OBSOLETE : XDrawPoint() slows down loading...
 *-----------------------------------------------------------------------------
 */

static int 
MyX_LoadPixmap_Direct (MyX * myx, char const filename [], 
                       Pixmap * img, Pixmap * msk, UInt * w, UInt * h)
{
#define MAX_ROW 10000
  UChar row [4 * MAX_ROW];
  MagickWand * wand;

  char * mode [2]= { "RGB", "RGBA" };
  int width, height, alpha;
  int i, j, j_src;
  ULong img_pix, msk_pix;

  if (img) * img= None;
  if (msk) * msk= None;
  if ((wand= NewMagickWand ()) == NULL) return 1;

  if (MagickReadImage(wand, filename) == MagickFalse) {
    DestroyMagickWand (wand); 
    return 2;
  }
  
  alpha= !!msk; /* MagickGetImageMatte  (wand); */ 
  
  height= MagickGetImageHeight (wand);
  width=  MagickGetImageWidth  (wand);
  if (width > MAX_ROW) {
    DestroyMagickWand (wand); 
    return 3;
  }
  if (img) * img= XCreatePixmap (myx->display, myx->root, 
                                 width, height, myx->depth);
  if (msk) * msk= XCreatePixmap (myx->display, myx->root,
                                 width, height, myx->depth);

  for (i= 0; i < height; i++) {
    MagickExportImagePixels (wand, 0,i, width,1, mode[alpha], CharPixel, row);
    for (j= 0, j_src= 0; j < width; j++, j_src += 3+alpha) {
      if (! alpha || row [j_src+3]) {
        img_pix= MyX_RGB (myx, row [j_src+0], row [j_src+1], row [j_src+2]);
        msk_pix= 0; /* black contour  */
      } else { 
        img_pix=  0; /* black shape   */
        msk_pix= ~0; /* white contour */
      }
      if (img) {
        XSetForeground (myx->display, myx->gc_sys, img_pix);
        XDrawPoint (myx->display, * img, myx->gc_sys, j, i);
      }
      if (msk) {
        XSetForeground (myx->display, myx->gc_sys, msk_pix);
        XDrawPoint (myx->display, * msk, myx->gc_sys, j, i);
      }
    }
  }
  DestroyMagickWand (wand);
  * w= width; * h= height;
  return 0;
#undef MAX_ROW
}

/*-----------------------------------------------------------------------------
 *
 *-----------------------------------------------------------------------------
 */

static int
MyX_LoadPixmap_ViaImage (MyX * myx, char const filename [], 
                         Pixmap * img, Pixmap * msk, UInt * w, UInt * h)
{
  XImage * ximg, * xmsk;
  int err= MyX_LoadImage (myx, filename, 
                          img ? & ximg : NULL, 
                          msk ? & xmsk : NULL, w, h);
  if (err) return err;
  if (img) {
    * img= XCreatePixmap (myx->display, myx->root, * w, * h, myx->depth);
    XPutImage (myx->display, *img, myx->gc_sys, ximg, 0, 0, 0, 0, * w, * h);
    XDestroyImage (ximg);
  }
  if (msk) {
    * msk= XCreatePixmap (myx->display, myx->root, * w, * h, myx->depth);
    XPutImage (myx->display, *msk, myx->gc_sys, xmsk, 0, 0, 0, 0, * w, * h);
    XDestroyImage (xmsk);
  }
 
  return err;
}

/*=============================================================================
 *
 *=============================================================================
 */

int
MyX_LoadPixmap (MyX * myx, char const filename [], 
                Pixmap * img, Pixmap * msk, UInt * w, UInt * h)
{
  int fast= 1;
  return fast ?
    MyX_LoadPixmap_ViaImage (myx, filename, img, msk, w, h) :
    MyX_LoadPixmap_Direct   (myx, filename, img, msk, w, h) ;
}
 
/*=============================================================================
 *
 *=============================================================================
 */

int
MyX_LoadSprite (MyX * myx, char const filename [], MyX_Sprite * s, int transp)
{
  s->x_flip= 0;
  s->y_flip= 0;
  
  if (! transp) s->mask= None;
  return MyX_LoadPixmap (myx, filename, 
                         & s->image, transp ? & s->mask : NULL, 
                         & s->width, & s->height);
}

 
/*=============================================================================
 *
 *=============================================================================
 */


void
MyX_FreeSprite (MyX * myx, MyX_Sprite * s) {
  if (s->image != None) XFreePixmap (myx->display, s->image);
  if (s->mask  != None) XFreePixmap (myx->display, s->mask);
}


/*=============================================================================
 *
 *=============================================================================
 */

void 
MyX_PutImage (MyX * myx, Drawable dst, GC gc,
              XImage * img, int x_dst, int y_dst)
{
  int x_src, w_src= img-> width; 
  int y_src, h_src= img->height; 
  MyX_Geometry g;

  if (img == NULL) return;

  if (x_dst < 0) { 
    w_src+= x_dst; if (w_src < 0) return;
    x_src= -x_dst; x_dst= 0;
  } else x_src=  0;

  if (y_dst < 0) { 
    h_src+= y_dst; if (h_src < 0) return;
    y_src= -y_dst; y_dst= 0;
  } else y_src=  0;

  MyX_GetGeometry (myx, dst, & g);

  if (x_dst + w_src > (int) g.w) w_src= g.w - x_dst;
  if (y_dst + h_src > (int) g.h) h_src= g.h - y_dst;

  XPutImage (myx->display, dst, gc, img,
             x_src, y_src, x_dst, y_dst, w_src, h_src);
}

/*=============================================================================
 *
 *=============================================================================
 */

void 
MyX_PutPixmap (MyX * myx, Drawable dst, GC gc, Drawable img, int x, int y)
{
  MyX_Geometry g;

  if (img == None) return;
  MyX_GetGeometry (myx, img, & g);
  XCopyArea (myx->display, img, dst, gc, 0, 0, g.w, g.h, x, y);
}

/*-----------------------------------------------------------------------------
 *
 *-----------------------------------------------------------------------------
 */

INLINE static void
MyX_XFlipImage_8bpp (XImage * img) 
{
  char * line;
  char * p1, * p2, p; 
  int y, useful_bytes_per_line= 1*img->width; /* all but padding ! */
  line= img->data;
  for (y= 0 ;  y < img->height; y++, line += img->bytes_per_line) {
    p2= (char *) (line + useful_bytes_per_line) - 1; 
    for (p1= (char *) line; p1 < p2; p1++, p2--) {
      p= * p1; * p1= * p2; * p2= p;
    }
  }
}

/*-----------------------------------------------------------------------------
 *
 *-----------------------------------------------------------------------------
 */

INLINE static void
MyX_XFlipImage_16bpp (XImage * img) 
{
  char * line;
  short * p1, * p2, p; 
  int y, useful_bytes_per_line= 2*img->width; /* all but padding ! */
  line= img->data;
  for (y= 0 ;  y < img->height; y++, line += img->bytes_per_line) {
    p2= (short *)(line + useful_bytes_per_line) - 1;
    for (p1= (short *) line; p1 < p2; p1++, p2--) {
      p= * p1; * p1= * p2; * p2= p;
    }
  }
}

/*-----------------------------------------------------------------------------
 *
 *-----------------------------------------------------------------------------
 */

INLINE static void
MyX_XFlipImage_24bpp_32bpp (XImage * img) 
{
  char * line;
  int * p1, * p2, p;
  int y, useful_bytes_per_line= 4*img->width; /* all but padding ! */
  line= img->data;
  for (y= 0 ;  y < img->height; y++, line += img->bytes_per_line) {
    p2= (int *) (line + useful_bytes_per_line) - 1;
    for (p1= (int *) line; p1 < p2; p1++, p2--) {
      p= * p1; * p1= * p2; * p2= p;
    }
  }
}

/*=============================================================================
 *
 *=============================================================================
 */

void
MyX_XFlipImage (XImage * img) 
{
  if (! img) return; 
  switch (img->bits_per_pixel) {
  case  8: MyX_XFlipImage_8bpp        (img); break;
  case 16: MyX_XFlipImage_16bpp       (img); break;
  default: MyX_XFlipImage_24bpp_32bpp (img); break;
  }
}

/*-----------------------------------------------------------------------------
 *
 *-----------------------------------------------------------------------------
 */

INLINE static void
MyX_YFlipImage_8bpp (XImage * img) 
{
  int x, y1, y2; 
  char * line1, * line2;
  char * p1, * p2, p; 
  line1= img->data; 
  line2= img->data + (img->height - 1) * img->bytes_per_line; 
  for (y1= 0, y2= img->height - 1;  y1 < y2; y1++, y2--) {
    p1= (char *) line1; 
    p2= (char *) line2; 
    for (x= 0; x < img->width; p1++, p2++, x++) {
      p= * p1; * p1= * p2; * p2= p;
    }
    line1 += img->bytes_per_line;
    line2 -= img->bytes_per_line;
  }
}

/*-----------------------------------------------------------------------------
 *
 *-----------------------------------------------------------------------------
 */

INLINE static void
MyX_YFlipImage_16bpp (XImage * img) 
{
  int x, y1, y2; 
  char * line1, * line2;
  short * p1, * p2, p; 
  line1= img->data; 
  line2= img->data + (img->height - 1) * img->bytes_per_line; 
  for (y1= 0, y2= img->height - 1;  y1 < y2; y1++, y2--) {
    p1= (short *) line1; 
    p2= (short *) line2; 
    for (x= 0; x < img->width; p1++, p2++, x++) {
      p= * p1; * p1= * p2; * p2= p;
    }
    line1 += img->bytes_per_line;
    line2 -= img->bytes_per_line;
  }
}

/*-----------------------------------------------------------------------------
 *
 *-----------------------------------------------------------------------------
 */

INLINE static void
MyX_YFlipImage_24bpp_32bpp (XImage * img) 
{
  int x, y1, y2; 
  char * line1, * line2;
  int * p1, * p2, p; 
  line1= img->data; 
  line2= img->data + (img->height - 1) * img->bytes_per_line; 
  for (y1= 0, y2= img->height - 1;  y1 < y2; y1++, y2--) {
    p1= (int *) line1; 
    p2= (int *) line2; 
    for (x= 0; x < img->width; p1++, p2++, x++) {
      p= * p1; * p1= * p2; * p2= p;
    }
    line1 += img->bytes_per_line;
    line2 -= img->bytes_per_line;
  }
}

/*=============================================================================
 *
 *=============================================================================
 */

void
MyX_YFlipImage (XImage * img) 
{
  if (! img) return; 
  switch (img->bits_per_pixel) {
  case  8: MyX_YFlipImage_8bpp        (img); break; 
  case 16: MyX_YFlipImage_16bpp       (img); break; 
  default: MyX_YFlipImage_24bpp_32bpp (img); break; 
  }
}

/*-----------------------------------------------------------------------------
 *
 *-----------------------------------------------------------------------------
 */

INLINE static void
MyX_XYFlipImage_8bpp (XImage * img) 
{
  int x, y1, y2; 
  char * line1, * line2;
  char * p1, * p2, p; 
  int useful_bytes_per_line= 1*img->width; /* all but padding ! */
  line1= img->data; 
  line2= img->data + (img->height - 1) * img->bytes_per_line; 
  for (y1= 0, y2= img->height - 1;  y1 < y2; y1++, y2--) {
    p1= (char *) line1; 
    p2= (char *) (line2 + useful_bytes_per_line) - 1; 
    for (x= 0; x < img->width; p1++, p2--, x++) {
      p= * p1; * p1= * p2; * p2= p;
    }
    line1 += img->bytes_per_line;
    line2 -= img->bytes_per_line;
  }
}

/*-----------------------------------------------------------------------------
 *
 *-----------------------------------------------------------------------------
 */

INLINE static void
MyX_XYFlipImage_16bpp (XImage * img) 
{
  int x, y1, y2; 
  char * line1, * line2;
  short * p1, * p2, p; 
  int useful_bytes_per_line= 2*img->width; /* all but padding ! */
  line1= img->data; 
  line2= img->data + (img->height - 1) * img->bytes_per_line; 
  for (y1= 0, y2= img->height - 1;  y1 < y2; y1++, y2--) {
    p1= (short *) line1; 
    p2= (short *) (line2 + useful_bytes_per_line) - 1; 
    for (x= 0; x < img->width; p1++, p2--, x++) {
      p= * p1; * p1= * p2; * p2= p;
    }
    line1 += img->bytes_per_line;
    line2 -= img->bytes_per_line;
  }
}

/*-----------------------------------------------------------------------------
 *
 *-----------------------------------------------------------------------------
 */

INLINE static void
MyX_XYFlipImage_24bpp_32bpp (XImage * img) 
{
  int x, y1, y2; 
  char * line1, * line2;
  int * p1, * p2, p; 
  int useful_bytes_per_line= 4*img->width; /* all but padding ! */
  line1= img->data; 
  line2= img->data + (img->height - 1) * img->bytes_per_line; 
  for (y1= 0, y2= img->height - 1;  y1 < y2; y1++, y2--) {
    p1= (int *) line1; 
    p2= (int *) (line2 + useful_bytes_per_line) - 1; 
    for (x= 0; x < img->width; p1++, p2--, x++) {
      p= * p1; * p1= * p2; * p2= p;
    }
    line1 += img->bytes_per_line;
    line2 -= img->bytes_per_line;
  }
}

/*=============================================================================
 *
 *=============================================================================
 */

void
MyX_XYFlipImage (XImage * img) 
{
  if (! img) return; 
  switch (img->bits_per_pixel) {
  case  8: MyX_XYFlipImage_8bpp        (img); break;
  case 16: MyX_XYFlipImage_16bpp       (img); break;
  default: MyX_XYFlipImage_24bpp_32bpp (img); break;
  }
}

/*=============================================================================
 *
 *=============================================================================
 */

void 
MyX_XFlipPixmap (MyX* myx, Pixmap p)
{
  MyX_Geometry g;
  XImage * ximg;
  if (p == None) return; 
  MyX_GetGeometry (myx, p, & g);
  ximg = XGetImage (myx->display, p, 0, 0, g.w, g.h, AllPlanes, ZPixmap);
  MyX_XFlipImage (ximg);
  XPutImage (myx->display, p, myx->gc_sys, ximg, 0, 0, 0, 0, g.w, g.h);
  XDestroyImage  (ximg);  
}

/*=============================================================================
 *
 *=============================================================================
 */

void MyX_YFlipPixmap (MyX* myx, Pixmap p)
{
  MyX_Geometry g;
  XImage * ximg;
  if (p == None) return; 
  MyX_GetGeometry (myx, p, & g);
  ximg = XGetImage (myx->display, p, 0, 0, g.w, g.h, AllPlanes, ZPixmap);
  MyX_YFlipImage (ximg);
  XPutImage (myx->display, p, myx->gc_sys, ximg, 0, 0, 0, 0, g.w, g.h);
  XDestroyImage  (ximg);  
}

/*=============================================================================
 *
 *=============================================================================
 */

void MyX_XYFlipPixmap (MyX* myx, Pixmap p)
{
  MyX_Geometry g;
  XImage * ximg;
  if (p == None) return; 
  MyX_GetGeometry (myx, p, & g);
  ximg = XGetImage (myx->display, p, 0, 0, g.w, g.h, AllPlanes, ZPixmap);
  MyX_XYFlipImage (ximg);
  XPutImage (myx->display, p, myx->gc_sys, ximg, 0, 0, 0, 0, g.w, g.h);
  XDestroyImage  (ximg);  
}

/*=============================================================================
 *
 *=============================================================================
 */

void MyX_XFlipSprite (MyX* myx, MyX_Sprite * s, int x_flip)
{
  if (x_flip == s->x_flip) return;
  MyX_XFlipPixmap (myx, s->image);
  MyX_XFlipPixmap (myx, s->mask);
  s->x_flip= ! s->x_flip;
}

/*=============================================================================
 *
 *=============================================================================
 */

void MyX_YFlipSprite (MyX* myx, MyX_Sprite * s, int y_flip)
{
  if (y_flip == s->y_flip) return;
  MyX_YFlipPixmap (myx, s->image);
  MyX_YFlipPixmap (myx, s->mask);
  s->y_flip= ! s->y_flip;
}

/*=============================================================================
 *
 *=============================================================================
 */

void MyX_XYFlipSprite (MyX* myx, MyX_Sprite * s, int x_flip, int y_flip)
{
  if (x_flip == s->x_flip && y_flip == s->y_flip) return;
  if (x_flip == s->x_flip) {
    MyX_YFlipPixmap (myx, s->image);
    MyX_YFlipPixmap (myx, s->mask);
    s->y_flip= ! s->y_flip;
    return;
  }
  if (y_flip == s->y_flip) {
    MyX_XFlipPixmap (myx, s->image);
    MyX_XFlipPixmap (myx, s->mask);
    s->x_flip= ! s->x_flip;
    return;
  }
  MyX_XYFlipPixmap (myx, s->image);
  MyX_XYFlipPixmap (myx, s->mask);
  s->x_flip= ! s->x_flip;
  s->y_flip= ! s->y_flip;
}

/*-----------------------------------------------------------------------------
 *
 *-----------------------------------------------------------------------------
 */

INLINE static void
MyX_PutSprite_Src (MyX * myx,  Drawable d, GC gc,  
                    MyX_Sprite * s, int x, int y) 
{
  XSetFunction (myx->display, gc, GXand);
  MyX_PutPixmap (myx, d, gc, s->mask, x, y);
  XSetFunction (myx->display, gc, GXor);
  MyX_PutPixmap (myx, d, gc, s->image, x, y);
  XSetFunction (myx->display, gc, GXcopy); 
}

INLINE static void
MyX_PutSprite_NotSrc (MyX * myx,  Drawable d, GC gc,  
                       MyX_Sprite * s, int x, int y) 
{
  XSetFunction (myx->display, gc, GXorInverted);
  MyX_PutPixmap (myx, d, gc, s->mask, x, y);

  XSetFunction (myx->display, gc, GXandInverted);
  MyX_PutPixmap (myx, d, gc, s->image, x, y);

  XSetFunction (myx->display, gc, GXcopy); 
}

INLINE static void
MyX_PutSprite_SrcOrDst (MyX * myx, Drawable d, GC gc,   
                         MyX_Sprite * s, int x, int y) 
{
  XSetFunction (myx->display, gc, GXor);
  MyX_PutPixmap (myx, d, gc, s->image, x, y);

  XSetFunction (myx->display, gc, GXcopy); 
}

INLINE static void
MyX_PutSprite_Black (MyX * myx, Drawable d, GC gc,   
                      MyX_Sprite * s, int x, int y) 
{
  XSetFunction (myx->display, gc, GXand);
  MyX_PutPixmap (myx, d, gc, s->mask, x, y);
  
  XSetFunction (myx->display, gc, GXcopy); 
}

INLINE static void
MyX_PutSprite_White (MyX * myx, Drawable d, GC gc,   
                      MyX_Sprite * s, int x, int y) 
{
  XSetFunction (myx->display, gc, GXorInverted);
  MyX_PutPixmap (myx, d, gc, s->mask, x, y);

  XSetFunction (myx->display, gc, GXcopy); 
}

INLINE static void
MyX_PutSprite_NotDst (MyX * myx, Drawable d, GC gc,  
                      MyX_Sprite * s, int x, int y) 
{
  XSetFunction (myx->display, gc, GXequiv);
  MyX_PutPixmap (myx, d, gc, s->mask, x, y);

  XSetFunction (myx->display, gc, GXcopy); 
}

INLINE static void
MyX_PutSprite_SrcXorDst (MyX * myx, Drawable d, GC gc,  
                         MyX_Sprite * s, int x, int y) 
{
  XSetFunction (myx->display, gc, GXxor);
  MyX_PutPixmap (myx, d, gc, s->image, x, y);

  XSetFunction (myx->display, gc, GXcopy); 
}

INLINE static void
MyX_PutSprite_NotSrcAndDst (MyX * myx, Drawable d, GC gc,  
                            MyX_Sprite * s, int x, int y) 
{
  XSetFunction (myx->display, gc, GXandInverted);
  MyX_PutPixmap (myx, d, gc, s->image, x, y);

  XSetFunction (myx->display, gc, GXcopy); 
}

INLINE static void
MyX_PutSprite_SrcNorDst (MyX * myx, Drawable d, GC gc,  
                         MyX_Sprite * s, int x, int y) 
{
  XSetFunction (myx->display, gc, GXequiv);
  MyX_PutPixmap (myx, d, gc, s->mask, x, y);

  XSetFunction (myx->display, gc, GXandInverted);
  MyX_PutPixmap (myx, d, gc, s->image, x, y);

  XSetFunction (myx->display, gc, GXcopy); 
}

INLINE static void
MyX_PutSprite_SrcOrNotDst (MyX * myx, Drawable d, GC gc,  
                           MyX_Sprite * s, int x, int y) 
{
  XSetFunction (myx->display, gc, GXequiv);
  MyX_PutPixmap (myx, d, gc, s->mask, x, y);

  XSetFunction (myx->display, gc, GXor);
  MyX_PutPixmap (myx, d, gc, s->image, x, y);

  XSetFunction (myx->display, gc, GXcopy); 
}

INLINE static void
MyX_PutSprite_SrcEqvDst (MyX * myx, Drawable d, GC gc,  
                         MyX_Sprite * s, int x, int y) 
{
  XSetFunction (myx->display, gc, GXequiv);
  MyX_PutPixmap (myx, d, gc, s->mask, x, y);

  XSetFunction (myx->display, gc, GXxor);
  MyX_PutPixmap (myx, d, gc, s->image, x, y);

  XSetFunction (myx->display, gc, GXcopy); 
}

INLINE static void
MyX_PutSprite_SrcAndDst (MyX * myx, Drawable d, GC gc, 
                          MyX_Sprite * s, int x, int y) 
{
  XSetFunction (myx->display, gc, GXxor);
  MyX_PutPixmap (myx, s->image, gc, s->mask, 0, 0);

  XSetFunction (myx->display, gc, GXand);
  MyX_PutPixmap (myx, d, gc, s->image, x, y);

  XSetFunction (myx->display, gc, GXxor);
  MyX_PutPixmap (myx, s->image, gc, s->mask, 0, 0);

  XSetFunction (myx->display, gc, GXcopy); 
}

INLINE static void
MyX_PutSprite_NotSrcOrDst (MyX * myx, Drawable d, GC gc,  
                            MyX_Sprite * s, int x, int y) 
{
  XSetFunction (myx->display, gc, GXequiv);
  MyX_PutPixmap (myx, s->image, gc, s->mask, 0, 0);

  XSetFunction (myx->display, gc, GXor);
  MyX_PutPixmap (myx, d, gc, s->image, x, y);

  XSetFunction (myx->display, gc, GXequiv);
  MyX_PutPixmap (myx, s->image, gc, s->mask, 0, 0);

  XSetFunction (myx->display, gc, GXcopy); 
}

INLINE static void
MyX_PutSprite_SrcNandDst (MyX * myx, Drawable d, GC gc,  
                           MyX_Sprite * s, int x, int y) 
{
  XSetFunction (myx->display, gc, GXequiv);
  MyX_PutPixmap (myx, s->image, gc, s->mask, 0, 0);

  XSetFunction (myx->display, gc, GXequiv);
  MyX_PutPixmap (myx, d, gc, s->mask, x, y);

  XSetFunction (myx->display, gc, GXor);
  MyX_PutPixmap (myx, d, gc, s->image, x, y);

  XSetFunction (myx->display, gc, GXequiv);
  MyX_PutPixmap (myx, s->image, gc, s->mask, 0, 0);

  XSetFunction (myx->display, gc, GXcopy); 
}

INLINE static void
MyX_PutSprite_SrcAndNotDst (MyX * myx, Drawable d, GC gc,  
                             MyX_Sprite * s, int x, int y) 
{
  XSetFunction (myx->display, gc, GXxor);
  MyX_PutPixmap (myx, s->image, gc, s->mask, 0, 0);

  XSetFunction (myx->display, gc, GXxor);
  MyX_PutPixmap (myx, d, gc, s->mask, x, y);

  XSetFunction (myx->display, gc, GXandReverse);
  MyX_PutPixmap (myx, d, gc, s->image, x, y);

  XSetFunction (myx->display, gc, GXxor);
  MyX_PutPixmap (myx, s->image, gc, s->mask, 0, 0);

  XSetFunction (myx->display, gc, GXcopy); 
}

/*=============================================================================
 *
 *=============================================================================
 */

void
MyX_PutSprite (MyX * myx, Drawable d, GC gc,
                MyX_Sprite * s, int x, int y) 
{
  if (s->image == None) return;
  if (s->mask  == None) 
    MyX_PutPixmap (myx, d, gc, s->image, x, y); 
  else
    MyX_PutSprite_Src (myx, d, gc, s, x, y);
}

/*=============================================================================
 *
 *=============================================================================
 */

void
MyX_PutFxSprite (MyX * myx, Drawable d, GC gc, UInt fx, 
                 MyX_Sprite * s, int x, int y) 
{
  if (s->image == None) return;
  if (s->mask == None) {
    XSetFunction (myx->display, gc, fx);
    MyX_PutPixmap (myx, d, gc, s->image, x, y); 
    XSetFunction (myx->display, gc, GXcopy);
    return;
  }
  switch (fx) {
  case  0: MyX_PutSprite_Black        (myx, d, gc, s, x, y); break;
  case  1: MyX_PutSprite_SrcAndDst    (myx, d, gc, s, x, y); break;
  case  2: MyX_PutSprite_SrcAndNotDst (myx, d, gc, s, x, y); break;
  case  3: MyX_PutSprite_Src          (myx, d, gc, s, x, y); break;
  case  4: MyX_PutSprite_NotSrcAndDst (myx, d, gc, s, x, y); break;
  case  5: break;
  case  6: MyX_PutSprite_SrcXorDst    (myx, d, gc, s, x, y); break;
  case  7: MyX_PutSprite_SrcOrDst     (myx, d, gc, s, x, y); break;
  case  8: MyX_PutSprite_SrcNorDst    (myx, d, gc, s, x, y); break;
  case  9: MyX_PutSprite_SrcEqvDst    (myx, d, gc, s, x, y); break;
  case 10: MyX_PutSprite_NotDst       (myx, d, gc, s, x, y); break;
  case 11: MyX_PutSprite_SrcOrNotDst  (myx, d, gc, s, x, y); break;
  case 12: MyX_PutSprite_NotSrc       (myx, d, gc, s, x, y); break;
  case 13: MyX_PutSprite_NotSrcOrDst  (myx, d, gc, s, x, y); break;
  case 14: MyX_PutSprite_SrcNandDst   (myx, d, gc, s, x, y); break;
  case 15: MyX_PutSprite_White        (myx, d, gc, s, x, y); break;
  default: break;
  }
}


