
#ifndef MYX_H___
#define MYX_H___

#include <stdlib.h>
#include <stdio.h>
#include <stdarg.h>

#include <math.h>
#include <string.h>
#include <locale.h>
#include <limits.h>
#include <time.h>

#include <unistd.h>
#include <sys/select.h>
#include <sys/time.h>
#include <sys/types.h>
#include <dirent.h>

#include <X11/X.h>
#include <X11/Xlib.h>
#include <X11/Xresource.h>
#include <X11/Xutil.h>
#include <X11/XKBlib.h>

/*
#ifdef __STRICT_ANSI__
#  define inline __inline__
#endif
#define INLINE inline
*/
#define INLINE

#define UNUSED  __attribute__ ((unused))

typedef unsigned char  UChar;
typedef unsigned long  ULong;
typedef unsigned int   UInt;
typedef unsigned short UShort;

extern const char * MyX_EventString [];

typedef struct MyX_Channel {
  ULong mask, shift, length, max;
} MyX_Channel;

typedef struct MyX_TrueColor 
{
  MyX_Channel green;
  MyX_Channel red;
  MyX_Channel blue;
} MyX_TrueColor;

typedef struct MyX_PseudoColor 
{
  Colormap colormap;
  ULong palette [8][8][4];
  XColor samples [256];
} MyX_PseudoColor;


typedef struct MyX MyX;

typedef void (* MyX_TimerHandler) (MyX * myx, void * data,
				   struct timeval * expected);

typedef struct MyX_Timer 
{
  struct timeval expected;
  MyX_TimerHandler handler;
  void * data;
} MyX_Timer;

#define MYX_MAX_TIMERS 10000

typedef struct MyX_TimerQueue {
  ULong nb_timers;
  MyX_Timer timers [MYX_MAX_TIMERS];
} MyX_TimerQueue;


struct MyX {
  Display * display; 
  Window root; 
  int screen;
  GC gc, gc_sys; 
  XFontStruct * font_fixed, * font_variable;
  int do_loop; 
  Visual * visual;
  UInt depth;
  MyX_PseudoColor pseudoColor;
  MyX_TrueColor   trueColor;   
  XContext dataContext, handlersContext;
  MyX_TimerQueue timerQueue;
  int ignoreAutoRepeat;
};

typedef void (* MyX_Handler) (MyX * myx, XEvent * event, void * data);

#define MYX_MAX_EVENTS LASTEvent

typedef struct MyX_Handlers {
  MyX_Handler table [MYX_MAX_EVENTS];
} MyX_Handlers;

typedef struct MyX_Sprite 
{
  Pixmap image, mask;
  UInt width, height;
  int x_flip, y_flip;
} MyX_Sprite;

typedef struct MyX_Geometry {
  Window root;
  int x, y;
  UInt w, h;
  UInt border, depth;
} MyX_Geometry;

typedef struct MyX_Tree {
  Window root, parent, * children;
  UInt nb_children;
} MyX_Tree;

int  MyX_Init  (MyX * myx, char const displayName []);
int  MyX_InitColor (MyX * myx);
void MyX_Clean (MyX * myx);

int MyX_LoadImage (MyX * myx, char const filename [], 
		   XImage ** img, XImage ** msk,
		   UInt * width, UInt * height);

void MyX_PutImage  (MyX* myx, Drawable d, GC gc, XImage* img, int x, int y);

void MyX_XFlipImage (XImage * img);
void MyX_YFlipImage (XImage * img);
void MyX_XYFlipImage(XImage * img); 

int MyX_LoadPixmap (MyX * myx, char const filename [], 
		    Pixmap * img, Pixmap * msk,
		    UInt * width, UInt * height);

void MyX_PutPixmap (MyX * myx, Drawable d, GC gc, Pixmap p, int x, int y);

void MyX_XFlipPixmap (MyX * myx, Pixmap p);
void MyX_YFlipPixmap (MyX * myx, Pixmap p);
void MyX_XYFlipPixmap(MyX * myx, Pixmap p);

int  MyX_LoadSprite (MyX * myx, char const filename [], MyX_Sprite * s, int transp);
void MyX_FreeSprite (MyX * myx, MyX_Sprite * s);

void MyX_PutSprite  (MyX* myx, Drawable  d, GC gc,  
		     MyX_Sprite * s, int x, int y);

void MyX_PutFxSprite (MyX * myx, Drawable d, GC gc, UInt fx, 
		      MyX_Sprite * s, int x, int y); 

void MyX_XFlipSprite (MyX * myx, MyX_Sprite * s, int x_flip);
void MyX_YFlipSprite (MyX * myx, MyX_Sprite * s, int y_flip);
void MyX_XYFlipSprite(MyX * myx, MyX_Sprite * s, int x_flip, int y_flip);


INLINE double MyRound (double x);
INLINE ULong MyX_RGB  (MyX * myx, UChar  r, UChar  g, UChar  b);
INLINE ULong MyX_RGBF (MyX * myx, double r, double g, double b);
INLINE ULong MyX_AverageColors (MyX * myx, ULong p1, ULong p2);
INLINE ULong MyX_BlendColors   (MyX * myx, ULong p1, double alpha, ULong p2);

int MyX_CallWithinDelay (MyX * myx, MyX_TimerHandler handler, void * data, 
			 struct timeval * delay, struct timeval * from);


void MyX_WaitEvents (MyX * myx, struct timeval * max_delay); 

void MyX_EventLoop (MyX * myx);

void MyX_IgnoreAutoRepeat (MyX * myx, int ignoreAutoRepeat);

void MyX_InitHandlers (MyX_Handlers * hanlders);

void MyX_SetHandler (MyX_Handlers * handlers, 
		     int eventType, MyX_Handler handler);

Window MyX_CreateWindow (MyX * myx, MyX_Handlers * handlers, void * data,
			 Window parent, int x, int y, UInt w, UInt h);

void MyX_GetGeometry (MyX * myx, Drawable d, MyX_Geometry * g);
void MyX_QueryTree   (MyX * myx, Window w, MyX_Tree * t);

void MyX_DestroyWindowContext (MyX * myx, Window window);

void MyX_SetWindowHandlers (MyX * myx, Window w, MyX_Handlers * h);
MyX_Handlers * MyX_GetWindowHandlers (MyX * myx, Window window);
void MyX_SetWindowData (MyX * myx, Window window, void * data);
void * MyX_GetWindowData (MyX * myx, Window window);

#ifndef M_PI
#  define M_PI 3.14159265358979323846
#endif

#ifndef timerisset
#define timerisset(tvp)	((tvp)->tv_sec || (tvp)->tv_usec)
#endif

#ifndef timerclear
# define timerclear(tvp) ((tvp)->tv_sec = (tvp)->tv_usec = 0)
#endif

#ifndef timercmp
#define timercmp(a,b,CMP) 	      \
  (((a)->tv_sec == (b)->tv_sec) ?     \
   ((a)->tv_usec CMP (b)->tv_usec) :  \
   ((a)->tv_sec CMP (b)->tv_sec))
#endif

#ifndef timeradd
#define timeradd(a,b,result)		             \
  do {						     \
    (result)->tv_sec = (a)->tv_sec + (b)->tv_sec;    \
    (result)->tv_usec = (a)->tv_usec + (b)->tv_usec; \
    if ((result)->tv_usec >= 1000000)		     \
      {						     \
	++(result)->tv_sec;		             \
	(result)->tv_usec -= 1000000;	             \
      }					             \
  } while (0)
#endif

#ifndef timersub
#define timersub(a,b,result)			     \
  do {						     \
    (result)->tv_sec = (a)->tv_sec - (b)->tv_sec;    \
    (result)->tv_usec = (a)->tv_usec - (b)->tv_usec; \
    if ((result)->tv_usec < 0) {		     \
      --(result)->tv_sec;			     \
      (result)->tv_usec += 1000000;		     \
    }						     \
  } while (0)
#endif

#endif
